#  Copyright (c) "Neo4j"
#  Neo4j Sweden AB [https://neo4j.com]
#  #
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  #
#      https://www.apache.org/licenses/LICENSE-2.0
#  #
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
from __future__ import annotations

import logging
from abc import ABC, abstractmethod
from typing import Any, List, Optional, Sequence, Type, Union

from pydantic import BaseModel

from neo4j_graphrag.message_history import MessageHistory
from neo4j_graphrag.tool import Tool
from neo4j_graphrag.types import LLMMessage
from neo4j_graphrag.utils.rate_limit import (
    DEFAULT_RATE_LIMIT_HANDLER,
    RateLimitHandler,
)

from .types import LLMResponse, ToolCallResponse

# pylint: disable=redefined-builtin

logger = logging.getLogger(__name__)


class LLMInterface(ABC):
    """Interface for large language models.

    Args:
        model_name (str): The name of the language model.
        model_params (Optional[dict]): Additional parameters passed to the model when text is sent to it. Defaults to None.
        rate_limit_handler (Optional[RateLimitHandler]): Handler for rate limiting. Defaults to retry with exponential backoff.
        **kwargs (Any): Arguments passed to the model when for the class is initialised. Defaults to None.
    """

    supports_structured_output: bool = False
    """Whether this LLM supports structured output (response_format with Pydantic models or json schema)."""

    def __init__(
        self,
        model_name: str,
        model_params: Optional[dict[str, Any]] = None,
        rate_limit_handler: Optional[RateLimitHandler] = None,
        **kwargs: Any,
    ):
        logger.warning(
            "LLMInterface is deprecated and will be removed in future versions. Please use LLMInterfaceV2 instead."
        )
        self.model_name = model_name
        self.model_params = model_params or {}

        if rate_limit_handler is not None:
            self._rate_limit_handler = rate_limit_handler
        else:
            self._rate_limit_handler = DEFAULT_RATE_LIMIT_HANDLER

    @abstractmethod
    def invoke(
        self,
        input: str,
        message_history: Optional[Union[List[LLMMessage], MessageHistory]] = None,
        system_instruction: Optional[str] = None,
    ) -> LLMResponse:
        """Sends a text input to the LLM and retrieves a response.

        Args:
            input (str): Text sent to the LLM.
            message_history (Optional[Union[List[LLMMessage], MessageHistory]]): A collection previous messages,
                with each message having a specific role assigned.
            system_instruction (Optional[str]): An option to override the llm system message for this invocation.

        Returns:
            LLMResponse: The response from the LLM.

        Raises:
            LLMGenerationError: If anything goes wrong.
        """

    @abstractmethod
    async def ainvoke(
        self,
        input: str,
        message_history: Optional[Union[List[LLMMessage], MessageHistory]] = None,
        system_instruction: Optional[str] = None,
    ) -> LLMResponse:
        """Asynchronously sends a text input to the LLM and retrieves a response.

        Args:
            input (str): Text sent to the LLM.
            message_history (Optional[Union[List[LLMMessage], MessageHistory]]): A collection previous messages,
                with each message having a specific role assigned.
            system_instruction (Optional[str]): An option to override the llm system message for this invocation.

        Returns:
            LLMResponse: The response from the LLM.

        Raises:
            LLMGenerationError: If anything goes wrong.
        """

    def invoke_with_tools(
        self,
        input: str,
        tools: Sequence[Tool],
        message_history: Optional[Union[List[LLMMessage], MessageHistory]] = None,
        system_instruction: Optional[str] = None,
    ) -> ToolCallResponse:
        """Sends a text input to the LLM with tool definitions and retrieves a tool call response.

        This is a default implementation that should be overridden by LLM providers that support tool/function calling.

        Args:
            input (str): Text sent to the LLM.
            tools (Sequence[Tool]): Sequence of Tools for the LLM to choose from. Each LLM implementation should handle the conversion to its specific format.
            message_history (Optional[Union[List[LLMMessage], MessageHistory]]): A collection previous messages,
                with each message having a specific role assigned.
            system_instruction (Optional[str]): An option to override the llm system message for this invocation.

        Returns:
            ToolCallResponse: The response from the LLM containing a tool call.

        Raises:
            LLMGenerationError: If anything goes wrong.
            NotImplementedError: If the LLM provider does not support tool calling.
        """
        raise NotImplementedError("This LLM provider does not support tool calling.")

    async def ainvoke_with_tools(
        self,
        input: str,
        tools: Sequence[Tool],
        message_history: Optional[Union[List[LLMMessage], MessageHistory]] = None,
        system_instruction: Optional[str] = None,
    ) -> ToolCallResponse:
        """Asynchronously sends a text input to the LLM with tool definitions and retrieves a tool call response.

        This is a default implementation that should be overridden by LLM providers that support tool/function calling.

        Args:
            input (str): Text sent to the LLM.
            tools (Sequence[Tool]): Sequence of Tools for the LLM to choose from. Each LLM implementation should handle the conversion to its specific format.
            message_history (Optional[Union[List[LLMMessage], MessageHistory]]): A collection previous messages,
                with each message having a specific role assigned.
            system_instruction (Optional[str]): An option to override the llm system message for this invocation.

        Returns:
            ToolCallResponse: The response from the LLM containing a tool call.

        Raises:
            LLMGenerationError: If anything goes wrong.
            NotImplementedError: If the LLM provider does not support tool calling.
        """
        raise NotImplementedError("This LLM provider does not support tool calling.")


class LLMInterfaceV2(ABC):
    """Interface for large language models compatible with LangChain.

    Args:
        model_name (str): The name of the language model.
        model_params (Optional[dict]): Additional parameters passed to the model when text is sent to it. Defaults to None.
        rate_limit_handler (Optional[RateLimitHandler]): Handler for rate limiting. Defaults to retry with exponential backoff.
        **kwargs (Any): Arguments passed to the model when for the class is initialised. Defaults to None.
    """

    def __init__(
        self,
        model_name: str,
        model_params: Optional[dict[str, Any]] = None,
        rate_limit_handler: Optional[RateLimitHandler] = None,
        **kwargs: Any,
    ):
        self.model_name = model_name
        self.model_params = model_params or {}

        if rate_limit_handler is not None:
            self._rate_limit_handler = rate_limit_handler
        else:
            self._rate_limit_handler = DEFAULT_RATE_LIMIT_HANDLER

    @abstractmethod
    def invoke(
        self,
        input: List[LLMMessage],
        response_format: Optional[Union[Type[BaseModel], dict[str, Any]]] = None,
        **kwargs: Any,
    ) -> LLMResponse:
        """Sends a list of messages to the LLM and retrieves a response.

        Args:
            input (List[LLMMessage]): Text sent to the LLM as a list of LLMMessage objects.
            response_format (Optional[Union[Type[BaseModel], dict[str, Any]]]): Optional
                response format specification. Can be a Pydantic model class for structured
                output or a dict for provider-specific formats. Defaults to None.

        Returns:
            LLMResponse: The response from the LLM.

        Raises:
            LLMGenerationError: If anything goes wrong.
            NotImplementedError: If the LLM provider does not support structured output.
        """

    @abstractmethod
    async def ainvoke(
        self,
        input: List[LLMMessage],
        response_format: Optional[Union[Type[BaseModel], dict[str, Any]]] = None,
        **kwargs: Any,
    ) -> LLMResponse:
        """Asynchronously sends a list of messages to the LLM and retrieves a response.

        Args:
            input (List[LLMMessage]): List of messages sent to the LLM.
            response_format (Optional[Union[Type[BaseModel], dict[str, Any]]]): Optional
                response format specification. Can be a Pydantic model class for structured
                output or a dict for provider-specific formats. Defaults to None.

        Returns:
            LLMResponse: The response from the LLM.

        Raises:
            LLMGenerationError: If anything goes wrong.
            NotImplementedError: If the LLM provider does not support structured output.
        """



"""In this example, the pipeline is defined in a JSON file 'pipeline_config.json'.
According to the configuration file, some parameters will be read from the env vars
(Neo4j credentials and the OpenAI API key).
"""

import asyncio

## If env vars are in a .env file, uncomment:
## (requires pip install python-dotenv)
# from dotenv import load_dotenv
# load_dotenv()
# env vars manually set for testing:
import os
from pathlib import Path

from neo4j_graphrag.experimental.pipeline.config.runner import PipelineRunner
from neo4j_graphrag.experimental.pipeline.pipeline import PipelineResult

os.environ["NEO4J_URI"] = "bolt://localhost:7687"
os.environ["NEO4J_USER"] = "neo4j"
os.environ["NEO4J_PASSWORD"] = "password"
# os.environ["OPENAI_API_KEY"] = "sk-..."


root_dir = Path(__file__).parent
# file_path = root_dir / "pipeline_config.json"
file_path = root_dir / "pipeline_config.yaml"

# Text to process
TEXT = """The son of Duke Leto Atreides and the Lady Jessica, Paul is the heir of House Atreides,
an aristocratic family that rules the planet Caladan, the rainy planet, since 10191."""


async def main() -> PipelineResult:
    pipeline = PipelineRunner.from_config_file(file_path)
    return await pipeline.run({"splitter": {"text": TEXT}})


if __name__ == "__main__":
    print(asyncio.run(main()))



"""This example demonstrates how to use SimpleKGPipeline with automatic schema extraction
from a PDF file. When no schema is provided to SimpleKGPipeline, automatic schema extraction
is performed using the LLM.

Note: This example requires an OpenAI API key to be set in the .env file.
"""

import asyncio
import logging
import os
from pathlib import Path
from dotenv import load_dotenv
import neo4j

from neo4j_graphrag.experimental.pipeline.kg_builder import SimpleKGPipeline
from neo4j_graphrag.llm import OpenAILLM
from neo4j_graphrag.embeddings import OpenAIEmbeddings

# Load environment variables from .env file
load_dotenv()

# Configure logging
logging.basicConfig()
logging.getLogger("neo4j_graphrag").setLevel(logging.INFO)

# PDF file path
root_dir = Path(__file__).parents[2]
PDF_FILE = str(
    root_dir / "data" / "Harry Potter and the Chamber of Secrets Summary.pdf"
)


async def run_kg_pipeline_with_auto_schema() -> None:
    """Run the SimpleKGPipeline with automatic schema extraction from a PDF file."""

    # Define Neo4j connection
    uri = os.getenv("NEO4J_URI", "neo4j://localhost:7687")
    user = os.getenv("NEO4J_USER", "neo4j")
    password = os.getenv("NEO4J_PASSWORD", "password")

    # Initialize the Neo4j driver
    driver = neo4j.GraphDatabase.driver(uri, auth=(user, password))

    # Create the LLM instance
    llm = OpenAILLM(model_name="gpt-5")

    # Create the embedder instance
    embedder = OpenAIEmbeddings()

    try:
        # Create a SimpleKGPipeline instance without providing a schema
        # This will trigger automatic schema extraction
        kg_builder = SimpleKGPipeline(
            llm=llm,
            driver=driver,
            embedder=embedder,
            from_pdf=True,
        )

        print(f"Processing PDF file: {PDF_FILE}")
        # Run the pipeline on the PDF file
        await kg_builder.run_async(file_path=PDF_FILE)

    finally:
        # Close connections
        await llm.async_client.close()
        driver.close()


async def main() -> None:
    """Run the example."""
    # Create data directory if it doesn't exist
    data_dir = root_dir / "data"
    data_dir.mkdir(exist_ok=True)

    # Check if the PDF file exists
    if not Path(PDF_FILE).exists():
        print(f"Warning: PDF file not found at {PDF_FILE}")
        print("Please replace with a valid PDF file path.")
        return

    # Run the pipeline
    await run_kg_pipeline_with_auto_schema()


if __name__ == "__main__":
    asyncio.run(main())



from unittest.mock import MagicMock

import pytest
from neo4j_graphrag.llm import LLMResponse
from neo4j_graphrag.retrievers import Text2CypherRetriever
from neo4j_graphrag.types import RetrieverResult, RetrieverResultItem


@pytest.mark.usefixtures("setup_neo4j_for_schema_query")
def test_t2c_retriever_search(driver: MagicMock, llm: MagicMock) -> None:
    t2c_query = """
    MATCH (a:LabelA {property_a: 'a'})-[:REL_TYPE]->(b:LabelB)
    RETURN a.property_a
    """
    retriever = Text2CypherRetriever(driver=driver, llm=llm)
    retriever.llm.invoke.return_value = LLMResponse(content=t2c_query)
    query_text = "dummy-text"
    results = retriever.search(query_text=query_text)
    assert isinstance(results, RetrieverResult)
    assert len(results.items) == 1
    for result in results.items:
        assert isinstance(result, RetrieverResultItem)
        assert "a.property_a" in result.content



#  Copyright (c) "Neo4j"
#  Neo4j Sweden AB [https://neo4j.com]
#  #
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  #
#      https://www.apache.org/licenses/LICENSE-2.0
#  #
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
from __future__ import annotations

import asyncio
import datetime
import tempfile
from typing import Sized
from unittest import mock
from unittest.mock import AsyncMock, call, patch

import pytest
from neo4j_graphrag.experimental.pipeline import Component, Pipeline
from neo4j_graphrag.experimental.pipeline.exceptions import PipelineDefinitionError
from neo4j_graphrag.experimental.pipeline.notification import (
    EventCallbackProtocol,
    EventType,
    PipelineEvent,
    TaskEvent,
    Event,
)
from neo4j_graphrag.experimental.pipeline.types.orchestration import RunResult

from .components import (
    ComponentAdd,
    ComponentMultiply,
    ComponentNoParam,
    ComponentPassThrough,
    StringResultModel,
    SlowComponentMultiply,
)


@pytest.mark.asyncio
async def test_simple_pipeline_two_components() -> None:
    pipe = Pipeline()
    component_a = ComponentNoParam()
    component_b = ComponentNoParam()
    pipe.add_component(
        component_a,
        "a",
    )
    pipe.add_component(
        component_b,
        "b",
    )
    pipe.connect("a", "b", {})
    with mock.patch(
        "tests.unit.experimental.pipeline.test_pipeline.ComponentNoParam.run"
    ) as mock_run:
        mock_run.side_effect = [
            StringResultModel(result="1"),
            StringResultModel(result="2"),
        ]
        res = await pipe.run({})
        mock_run.assert_awaited_with(**{})
        mock_run.assert_awaited_with(**{})
    assert "b" in res.result
    assert res.result["b"] == {"result": "2"}


@pytest.mark.asyncio
async def test_pipeline_parameter_propagation() -> None:
    pipe = Pipeline()
    component_a = ComponentPassThrough()
    component_b = ComponentPassThrough()
    pipe.add_component(component_a, "a")
    pipe.add_component(component_b, "b")
    # first component output product goes to second component input number1
    pipe.connect("a", "b", {"value": "a.result"})
    with mock.patch(
        "tests.unit.experimental.pipeline.test_pipeline.ComponentPassThrough.run"
    ) as mock_run:
        mock_run.side_effect = [
            StringResultModel(result="1"),
            StringResultModel(result="2"),
        ]
        res = await pipe.run({"a": {"value": "text"}})
        mock_run.assert_has_awaits([call(**{"value": "text"}), call(**{"value": "1"})])
    assert res.result == {"b": {"result": "2"}}


def test_pipeline_parameter_validation_no_expected_params() -> None:
    pipe = Pipeline()
    component_a = ComponentNoParam()
    pipe.add_component(component_a, "a")
    is_valid = pipe.validate_parameter_mapping_for_task(pipe.get_node_by_name("a"))
    assert is_valid is True


def test_pipeline_parameter_validation_one_component_all_good() -> None:
    pipe = Pipeline()
    component_a = ComponentPassThrough()
    pipe.add_component(component_a, "a")
    is_valid = pipe.validate_parameter_mapping_for_task(pipe.get_node_by_name("a"))
    assert is_valid is True


def test_pipeline_invalidate() -> None:
    pipe = Pipeline()
    pipe.is_validated = True
    pipe.param_mapping = {"a": {"key": {"component": "component", "param": "param"}}}
    pipe.missing_inputs = {"a": ["other_key"]}
    pipe.invalidate()
    assert pipe.is_validated is False
    assert len(pipe.param_mapping) == 0
    assert len(pipe.missing_inputs) == 0


def test_pipeline_parameter_validation_called_twice() -> None:
    pipe = Pipeline()
    component_a = ComponentPassThrough()
    component_b = ComponentPassThrough()
    pipe.add_component(component_a, "a")
    pipe.add_component(component_b, "b")
    pipe.connect("a", "b", {"value": "a.result"})
    is_valid = pipe.validate_parameter_mapping_for_task(pipe.get_node_by_name("b"))
    assert is_valid is True
    with pytest.raises(PipelineDefinitionError):
        pipe.validate_parameter_mapping_for_task(pipe.get_node_by_name("b"))
    pipe.invalidate()
    is_valid = pipe.validate_parameter_mapping_for_task(pipe.get_node_by_name("b"))
    assert is_valid is True


def test_pipeline_parameter_validation_one_component_input_param_missing() -> None:
    pipe = Pipeline()
    component_a = ComponentPassThrough()
    pipe.add_component(component_a, "a")
    pipe.validate_parameter_mapping_for_task(pipe.get_node_by_name("a"))
    assert pipe.missing_inputs["a"] == ["value"]


def test_pipeline_parameter_validation_param_mapped_twice() -> None:
    pipe = Pipeline()
    component_a = ComponentPassThrough()
    component_b = ComponentPassThrough()
    component_c = ComponentPassThrough()
    pipe.add_component(component_a, "a")
    pipe.add_component(component_b, "b")
    pipe.add_component(component_c, "c")
    pipe.connect("a", "c", {"value": "a.result"})
    pipe.connect("b", "c", {"value": "b.result"})
    with pytest.raises(PipelineDefinitionError) as excinfo:
        pipe.validate_parameter_mapping_for_task(pipe.get_node_by_name("c"))
        assert (
            "Parameter 'value' already mapped to {'component': 'a', 'param': 'result'}"
            in str(excinfo)
        )


def test_pipeline_parameter_validation_unexpected_input() -> None:
    pipe = Pipeline()
    component_a = ComponentPassThrough()
    component_b = ComponentPassThrough()
    pipe.add_component(component_a, "a")
    pipe.add_component(component_b, "b")
    pipe.connect("a", "b", {"unexpected_input_name": "a.result"})
    with pytest.raises(PipelineDefinitionError) as excinfo:
        pipe.validate_parameter_mapping_for_task(pipe.get_node_by_name("b"))
        assert (
            "Parameter 'unexpected_input_name' is not a valid input for component 'b' of type 'ComponentPassThrough'"
            in str(excinfo)
        )


def test_pipeline_parameter_validation_connected_components_input() -> None:
    """Parameter for component 'b' comes from the pipeline inputs"""
    pipe = Pipeline()
    component_a = ComponentNoParam()
    component_b = ComponentPassThrough()
    pipe.add_component(component_a, "a")
    pipe.add_component(component_b, "b")
    pipe.connect("a", "b", {})
    is_valid = pipe.validate_parameter_mapping_for_task(pipe.get_node_by_name("b"))
    assert is_valid is True
    assert dict(pipe.missing_inputs) == {"b": ["value"]}


def test_pipeline_parameter_validation_connected_components_result() -> None:
    """Parameter for component 'b' comes from the result of component 'a'"""
    pipe = Pipeline()
    component_a = ComponentNoParam()
    component_b = ComponentPassThrough()
    pipe.add_component(component_a, "a")
    pipe.add_component(component_b, "b")
    pipe.connect("a", "b", {"value": "b.result"})
    is_valid = pipe.validate_parameter_mapping_for_task(pipe.get_node_by_name("b"))
    assert is_valid is True
    assert pipe.missing_inputs == {"b": []}


def test_pipeline_parameter_validation_connected_components_missing_input() -> None:
    """Parameter for component 'b' is missing"""
    pipe = Pipeline()
    component_a = ComponentNoParam()
    component_b = ComponentPassThrough()
    pipe.add_component(component_a, "a")
    pipe.add_component(component_b, "b")
    pipe.connect("a", "b", {})
    is_valid = pipe.validate_parameter_mapping_for_task(pipe.get_node_by_name("b"))
    assert is_valid is True
    assert pipe.missing_inputs["b"] == ["value"]


def test_pipeline_parameter_validation_full_missing_inputs_in_user_data() -> None:
    pipe = Pipeline()
    component_a = ComponentNoParam()
    component_b = ComponentPassThrough()
    pipe.add_component(component_a, "a")
    pipe.add_component(component_b, "b")
    pipe.connect("a", "b", {})
    is_valid = pipe.validate_input_data(data={"b": {"value": "input for b"}})
    assert is_valid is True


def test_pipeline_parameter_validation_full_missing_inputs_in_component_name() -> None:
    pipe = Pipeline()
    component_a = ComponentNoParam()
    component_b = ComponentPassThrough()
    pipe.add_component(component_a, "a")
    pipe.add_component(component_b, "b")
    pipe.connect("a", "b", {})
    with pytest.raises(PipelineDefinitionError):
        pipe.validate_input_data(data={"b": {}})


def test_pipeline_parameter_validation_full_missing_inputs() -> None:
    pipe = Pipeline()
    component_a = ComponentNoParam()
    component_b = ComponentPassThrough()
    pipe.add_component(component_a, "a")
    pipe.add_component(component_b, "b")
    pipe.connect("a", "b", {})
    with pytest.raises(PipelineDefinitionError):
        pipe.validate_input_data(data={})


@pytest.mark.asyncio
async def test_pipeline_branches() -> None:
    pipe = Pipeline()
    component_a = AsyncMock(spec=Component)
    component_a.run_with_context = AsyncMock(return_value={})
    component_b = AsyncMock(spec=Component)
    component_b.run_with_context = AsyncMock(return_value={})
    component_c = AsyncMock(spec=Component)
    component_c.run_with_context = AsyncMock(return_value={})

    pipe.add_component(component_a, "a")
    pipe.add_component(component_b, "b")
    pipe.add_component(component_c, "c")
    pipe.connect("a", "b")
    pipe.connect("a", "c")
    pipeline_result = await pipe.run({})
    res = pipeline_result.result
    assert "b" in res
    assert "c" in res


@pytest.mark.asyncio
async def test_pipeline_aggregation() -> None:
    pipe = Pipeline()
    component_a = AsyncMock(spec=Component)
    component_a.run_with_context = AsyncMock(return_value={})
    component_b = AsyncMock(spec=Component)
    component_b.run_with_context = AsyncMock(return_value={})
    component_c = AsyncMock(spec=Component)
    component_c.run_with_context = AsyncMock(return_value={})

    pipe.add_component(
        component_a,
        "a",
    )
    pipe.add_component(
        component_b,
        "b",
    )
    pipe.add_component(component_c, "c")
    pipe.connect("a", "c")
    pipe.connect("b", "c")
    pipeline_result = await pipe.run({})
    res = pipeline_result.result
    assert "c" in res


@pytest.mark.asyncio
async def test_pipeline_missing_param_on_init() -> None:
    pipe = Pipeline()
    component_a = ComponentAdd()
    component_b = ComponentAdd()
    pipe.add_component(component_a, "a")
    pipe.add_component(component_b, "b")
    pipe.connect("a", "b", {"number1": "a.result"})
    with pytest.raises(PipelineDefinitionError) as excinfo:
        await pipe.run({"a": {"number1": 1}})
        assert (
            "Missing input parameters for a: Expected parameters: ['number1', 'number2']. Got: ['number1']"
            in str(excinfo.value)
        )


@pytest.mark.asyncio
async def test_pipeline_missing_param_on_connect() -> None:
    pipe = Pipeline()
    component_a = ComponentAdd()
    component_b = ComponentAdd()
    pipe.add_component(component_a, "a")
    pipe.add_component(component_b, "b")
    pipe.connect("a", "b", {"number1": "a.result"})
    with pytest.raises(PipelineDefinitionError) as excinfo:
        await pipe.run({"a": {"number1": 1, "number2": 2}})
        assert (
            "Missing input parameters for b: Expected parameters: ['number1', 'number2']. Got: ['number1']"
            in str(excinfo.value)
        )


@pytest.mark.asyncio
async def test_pipeline_with_default_params() -> None:
    pipe = Pipeline()
    component_a = ComponentAdd()
    component_b = ComponentMultiply()
    pipe.add_component(component_a, "a")
    pipe.add_component(component_b, "b")
    pipe.connect("a", "b", {"number1": "a.result"})
    pipeline_result = await pipe.run({"a": {"number1": 1, "number2": 2}})
    res = pipeline_result.result
    assert res == {"b": {"result": 6}}  # (1+2)*2


@pytest.mark.asyncio
async def test_pipeline_cycle() -> None:
    pipe = Pipeline()
    component_a = ComponentNoParam()
    component_b = ComponentNoParam()
    pipe.add_component(component_a, "a")
    pipe.add_component(component_b, "b")
    pipe.connect("a", "b", {})
    with pytest.raises(PipelineDefinitionError) as excinfo:
        pipe.connect("b", "a", {})
        assert "Cycles are not allowed" in str(excinfo.value)


@pytest.mark.asyncio
async def test_pipeline_wrong_component_name() -> None:
    pipe = Pipeline()
    component_a = ComponentNoParam()
    component_b = ComponentNoParam()
    pipe.add_component(component_a, "a")
    pipe.add_component(component_b, "b")
    with pytest.raises(PipelineDefinitionError) as excinfo:
        pipe.connect("a", "c", {})
        assert "a or c not in the Pipeline" in str(excinfo.value)


@pytest.mark.asyncio
async def test_pipeline_async() -> None:
    pipe = Pipeline()
    pipe.add_component(ComponentAdd(), "add")
    run_params = [[1, 20], [10, 2]]
    runs = []
    for a, b in run_params:
        runs.append(pipe.run({"add": {"number1": a, "number2": b}}))
    pipeline_result = await asyncio.gather(*runs)
    assert len(pipeline_result) == 2
    assert pipeline_result[0].run_id != pipeline_result[1].run_id
    assert pipeline_result[0].result == {"add": {"result": 21}}
    assert pipeline_result[1].result == {"add": {"result": 12}}


def test_pipeline_to_viz() -> None:
    pipe = Pipeline()
    component_a = ComponentAdd()
    component_b = ComponentMultiply()
    pipe.add_component(component_a, "a")
    pipe.add_component(component_b, "b")
    pipe.connect("a", "b", {"number1": "a.result"})
    g = pipe._get_neo4j_viz_graph()
    # 3 nodes:
    #   - 2 components 'a' and 'b'
    #   - 1 output 'a.result'
    assert len(g.nodes) == 3
    g = pipe._get_neo4j_viz_graph(hide_unused_outputs=False)
    # 4 nodes:
    #   - 2 components 'a' and 'b'
    #   - 2 output 'a.result' and 'b.result'
    assert len(g.nodes) == 4


def test_pipeline_draw() -> None:
    pipe = Pipeline()
    pipe.add_component(ComponentAdd(), "add")
    t = tempfile.NamedTemporaryFile(suffix=".html")
    pipe.draw(t.name)
    content = t.file.read()
    assert len(content) > 0


@patch("neo4j_graphrag.experimental.pipeline.pipeline.neo4j_viz_available", False)
def test_pipeline_draw_missing_neo4j_viz_dep() -> None:
    pipe = Pipeline()
    pipe.add_component(ComponentAdd(), "add")
    t = tempfile.NamedTemporaryFile(suffix=".html")
    with pytest.raises(ImportError):
        pipe.draw(t.name)


def test_run_result_no_warning(recwarn: Sized) -> None:
    RunResult()
    assert len(recwarn) == 0


@pytest.mark.asyncio
async def test_pipeline_event_notification() -> None:
    callback = AsyncMock(spec=EventCallbackProtocol)
    pipe = Pipeline(callback=callback)
    component_a = ComponentMultiply()
    pipe.add_component(
        component_a,
        "a",
    )
    a_input_data = {"number1": 2, "number2": 3}
    pipeline_result = await pipe.run({"a": a_input_data})

    await_calls = callback.await_args_list

    expected_event_list = [
        PipelineEvent(
            event_type=EventType.PIPELINE_STARTED,
            run_id=pipeline_result.run_id,
            timestamp=datetime.datetime.now(),
            message=None,
            payload={"a": a_input_data},
        ),
        TaskEvent(
            event_type=EventType.TASK_STARTED,
            run_id=pipeline_result.run_id,
            task_name="a",
            timestamp=datetime.datetime.now(),
            message=None,
            payload=a_input_data,
        ),
        TaskEvent(
            event_type=EventType.TASK_FINISHED,
            run_id=pipeline_result.run_id,
            task_name="a",
            timestamp=datetime.datetime.now(),
            message=None,
            payload={"result": 6},
        ),
        PipelineEvent(
            event_type=EventType.PIPELINE_FINISHED,
            run_id=pipeline_result.run_id,
            timestamp=datetime.datetime.now(),
            message=None,
            payload={"a": {"result": 6}},
        ),
    ]
    assert len(await_calls) == len(expected_event_list)

    previous_ts = None
    for await_call, expected_event in zip(await_calls, expected_event_list):
        actual_event = await_call[0][0]
        assert isinstance(actual_event, type(expected_event))
        assert actual_event.event_type == expected_event.event_type
        assert actual_event.run_id == expected_event.run_id
        assert actual_event.message == expected_event.message
        assert actual_event.payload == expected_event.payload
        if previous_ts:
            assert actual_event.timestamp > previous_ts
        previous_ts = actual_event.timestamp


@pytest.mark.asyncio
async def test_pipeline_event_notification_error_in_pipeline_run() -> None:
    callback = AsyncMock(spec=EventCallbackProtocol)
    pipe = Pipeline(callback=callback)
    component_a = ComponentAdd()
    component_b = ComponentAdd()
    pipe.add_component(component_a, "a")
    pipe.add_component(component_b, "b")
    pipe.connect("a", "b", {"number1": "a.result"})

    with pytest.raises(PipelineDefinitionError):
        await pipe.run({"a": {"number1": 1, "number2": 2}})
    assert len(callback.await_args_list) == 2
    assert callback.await_args_list[0][0][0].event_type == EventType.PIPELINE_STARTED
    assert callback.await_args_list[1][0][0].event_type == EventType.PIPELINE_FAILED


def test_event_model_no_warning(recwarn: Sized) -> None:
    event = Event(
        event_type=EventType.PIPELINE_STARTED,
        run_id="run_id",
        message=None,
        payload=None,
    )
    assert event.timestamp is not None
    assert len(recwarn) == 0


@pytest.mark.asyncio
async def test_pipeline_streaming_no_user_callback_happy_path() -> None:
    pipe = Pipeline()
    events = []
    async for e in pipe.stream({}):
        events.append(e)
    assert len(events) == 2
    assert events[0].event_type == EventType.PIPELINE_STARTED
    assert events[1].event_type == EventType.PIPELINE_FINISHED
    assert len(pipe.event_notifier.callbacks) == 0


@pytest.mark.asyncio
async def test_pipeline_streaming_with_user_callback_happy_path() -> None:
    callback = AsyncMock()
    pipe = Pipeline(callback=callback)
    events = []
    async for e in pipe.stream({}):
        events.append(e)
    assert len(events) == 2
    assert len(callback.call_args_list) == 2
    assert len(pipe.event_notifier.callbacks) == 1


@pytest.mark.asyncio
async def test_pipeline_streaming_very_long_running_user_callback() -> None:
    async def callback(event: Event) -> None:
        await asyncio.sleep(2)

    pipe = Pipeline(callback=callback)
    events = []
    async for e in pipe.stream({}):
        events.append(e)
    assert len(events) == 2
    assert len(pipe.event_notifier.callbacks) == 1


@pytest.mark.asyncio
async def test_pipeline_streaming_very_long_running_pipeline() -> None:
    slow_component = SlowComponentMultiply()
    pipe = Pipeline()
    pipe.add_component(slow_component, "slow_component")
    events = []
    async for e in pipe.stream({"slow_component": {"number1": 1, "number2": 2}}):
        events.append(e)
    assert len(events) == 4
    last_event = events[-1]
    assert last_event.event_type == EventType.PIPELINE_FINISHED
    assert last_event.payload == {"slow_component": {"result": 2}}


@pytest.mark.asyncio
async def test_pipeline_streaming_error_in_pipeline_definition() -> None:
    pipe = Pipeline()
    component_a = ComponentAdd()
    component_b = ComponentAdd()
    pipe.add_component(component_a, "a")
    pipe.add_component(component_b, "b")
    pipe.connect("a", "b", {"number1": "a.result"})
    events = []
    with pytest.raises(PipelineDefinitionError):
        async for e in pipe.stream({"a": {"number1": 1, "number2": 2}}):
            events.append(e)
    assert len(events) == 2
    assert events[0].event_type == EventType.PIPELINE_STARTED
    assert events[1].event_type == EventType.PIPELINE_FAILED


@pytest.mark.asyncio
async def test_pipeline_streaming_error_in_component() -> None:
    component = ComponentMultiply()
    pipe = Pipeline()
    pipe.add_component(component, "component")
    events = []
    with pytest.raises(TypeError):
        async for e in pipe.stream({"component": {"number1": None, "number2": 2}}):
            events.append(e)
    assert len(events) == 3
    assert events[0].event_type == EventType.PIPELINE_STARTED
    assert events[1].event_type == EventType.TASK_STARTED
    assert events[2].event_type == EventType.PIPELINE_FAILED


@pytest.mark.asyncio
async def test_pipeline_streaming_error_in_user_callback() -> None:
    async def callback(event: Event) -> None:
        raise Exception("error in callback")

    pipe = Pipeline(callback=callback)
    events = []
    async for e in pipe.stream({}):
        events.append(e)
    assert len(events) == 2
    assert len(pipe.event_notifier.callbacks) == 1



#  Copyright (c) "Neo4j"
#  Neo4j Sweden AB [https://neo4j.com]
#  #
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  #
#      https://www.apache.org/licenses/LICENSE-2.0
#  #
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
from __future__ import annotations

import asyncio

from neo4j_graphrag.embeddings.openai import OpenAIEmbeddings
from neo4j_graphrag.experimental.components.embedder import TextChunkEmbedder
from neo4j_graphrag.experimental.components.entity_relation_extractor import (
    LLMEntityRelationExtractor,
    OnError,
)
from neo4j_graphrag.experimental.components.kg_writer import Neo4jWriter
from neo4j_graphrag.experimental.components.schema import (
    SchemaBuilder,
    NodeType,
    PropertyType,
    RelationshipType,
)
from neo4j_graphrag.experimental.components.text_splitters.fixed_size_splitter import (
    FixedSizeSplitter,
)
from neo4j_graphrag.experimental.pipeline import Pipeline
from neo4j_graphrag.experimental.pipeline.pipeline import PipelineResult
from neo4j_graphrag.llm import LLMInterface, OpenAILLM

import neo4j


async def define_and_run_pipeline(
    neo4j_driver: neo4j.Driver, llm: LLMInterface
) -> PipelineResult:
    """This is where we define and run the KG builder pipeline, instantiating a few
    components:
    - Text Splitter: in this example we use the fixed size text splitter
    - Chunk Embedder: to embed the chunks' text
    - Schema Builder: this component takes a list of entities, relationships and
        possible triplets as inputs, validate them and return a schema ready to use
        for the rest of the pipeline
    - LLM Entity Relation Extractor is an LLM-based entity and relation extractor:
        based on the provided schema, the LLM will do its best to identity these
        entities and their relations within the provided text
    - KG writer: once entities and relations are extracted, they can be writen
        to a Neo4j database
    """
    pipe = Pipeline()
    # define the components
    pipe.add_component(
        # chunk_size=50 for the sake of this demo
        FixedSizeSplitter(chunk_size=4000, chunk_overlap=200, approximate=False),
        "splitter",
    )
    pipe.add_component(TextChunkEmbedder(embedder=OpenAIEmbeddings()), "chunk_embedder")
    pipe.add_component(SchemaBuilder(), "schema")
    pipe.add_component(
        LLMEntityRelationExtractor(
            llm=llm,
            on_error=OnError.RAISE,
        ),
        "extractor",
    )
    pipe.add_component(Neo4jWriter(neo4j_driver), "writer")
    # define the execution order of component
    # and how the output of previous components must be used
    pipe.connect("splitter", "chunk_embedder", input_config={"text_chunks": "splitter"})
    pipe.connect("schema", "extractor", input_config={"schema": "schema"})
    pipe.connect(
        "chunk_embedder", "extractor", input_config={"chunks": "chunk_embedder"}
    )
    pipe.connect(
        "extractor",
        "writer",
        input_config={"graph": "extractor"},
    )
    # user input:
    # the initial text
    # and the list of entities and relations we are looking for
    pipe_inputs = {
        "splitter": {
            "text": """Albert Einstein was a German physicist born in 1879 who
            wrote many groundbreaking papers especially about general relativity
            and quantum mechanics. He worked for many different institutions, including
            the University of Bern in Switzerland and the University of Oxford."""
        },
        "schema": {
            "node_types": [
                NodeType(
                    label="Person",
                    properties=[
                        PropertyType(name="name", type="STRING"),
                        PropertyType(name="place_of_birth", type="STRING"),
                        PropertyType(name="date_of_birth", type="DATE"),
                    ],
                ),
                NodeType(
                    label="Organization",
                    properties=[
                        PropertyType(name="name", type="STRING"),
                        PropertyType(name="country", type="STRING"),
                    ],
                ),
                NodeType(
                    label="Field",
                    properties=[
                        PropertyType(name="name", type="STRING"),
                    ],
                ),
            ],
            "relationship_types": [
                RelationshipType(
                    label="WORKED_ON",
                ),
                RelationshipType(
                    label="WORKED_FOR",
                ),
            ],
            "patterns": [
                ("Person", "WORKED_ON", "Field"),
                ("Person", "WORKED_FOR", "Organization"),
            ],
        },
        "extractor": {
            "document_info": {
                "path": "my text",
            }
        },
    }
    # run the pipeline
    return await pipe.run(pipe_inputs)


async def main() -> PipelineResult:
    llm = OpenAILLM(
        model_name="gpt-5",
        model_params={
            "max_tokens": 1000,
            "response_format": {"type": "json_object"},
        },
    )
    driver = neo4j.GraphDatabase.driver(
        "bolt://localhost:7687", auth=("neo4j", "password")
    )
    res = await define_and_run_pipeline(driver, llm)
    driver.close()
    await llm.async_client.close()
    return res


if __name__ == "__main__":
    res = asyncio.run(main())
    print(res)



#  Copyright (c) "Neo4j"
#  Neo4j Sweden AB [https://neo4j.com]
#  #
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  #
#      https://www.apache.org/licenses/LICENSE-2.0
#  #
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import pytest
from neo4j import Driver
from neo4j_graphrag.schema import (
    BASE_ENTITY_LABEL,
    EXCLUDED_LABELS,
    EXCLUDED_RELS,
    NODE_PROPERTIES_QUERY,
    REL_PROPERTIES_QUERY,
    REL_QUERY,
    query_database,
)


@pytest.mark.usefixtures("setup_neo4j_for_schema_query_with_excluded_labels")
def test_filtering_labels_node_properties(driver: Driver) -> None:
    node_properties = [
        data["output"]
        for data in query_database(
            driver,
            NODE_PROPERTIES_QUERY,
            params={"EXCLUDED_LABELS": EXCLUDED_LABELS, "SAMPLE": 1000},
        )
    ]

    assert node_properties == []


@pytest.mark.usefixtures("setup_neo4j_for_schema_query_with_excluded_labels")
def test_filtering_labels_relationship_properties(driver: Driver) -> None:
    relationship_properties = [
        data["output"]
        for data in query_database(
            driver,
            REL_PROPERTIES_QUERY,
            params={"EXCLUDED_LABELS": EXCLUDED_RELS, "SAMPLE": 1000},
        )
    ]

    assert relationship_properties == []


@pytest.mark.usefixtures("setup_neo4j_for_schema_query_with_excluded_labels")
def test_filtering_labels_relationships(driver: Driver) -> None:
    relationships = [
        data["output"]
        for data in query_database(
            driver,
            REL_QUERY,
            params={
                "EXCLUDED_LABELS": EXCLUDED_LABELS + [BASE_ENTITY_LABEL],
                "SAMPLE": 1000,
            },
        )
    ]

    assert relationships == []



"""This example demonstrates how to use PineconeNeo4jRetriever, ie vectors are
stored in the Pinecone database.

See the [README](./README.md) for more
information about how spin up a Pinecone and Neo4j databases if needed.

In this example, search is performed from a text. Embeddings are computed
using OpenAI models. See [../../embeddings/](../../embeddings/) for examples
using other supported embedders.
"""

from neo4j import GraphDatabase
from neo4j_graphrag.embeddings import OpenAIEmbeddings
from neo4j_graphrag.retrievers import PineconeNeo4jRetriever
from pinecone import Pinecone

NEO4J_AUTH = ("neo4j", "password")
NEO4J_URL = "neo4j://localhost:7687"
PC_API_KEY = "API_KEY"


def main() -> None:
    with GraphDatabase.driver(NEO4J_URL, auth=NEO4J_AUTH) as neo4j_driver:
        pc_client = Pinecone(PC_API_KEY)
        embedder = OpenAIEmbeddings()

        retriever = PineconeNeo4jRetriever(
            driver=neo4j_driver,
            client=pc_client,
            index_name="jeopardy",
            id_property_neo4j="id",
            embedder=embedder,
        )

        res = retriever.search(query_text="biology", top_k=2)
        print(res)


if __name__ == "__main__":
    main()



#  Copyright (c) "Neo4j"
#  Neo4j Sweden AB [https://neo4j.com]
#  #
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  #
#      https://www.apache.org/licenses/LICENSE-2.0
#  #
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
from pathlib import Path
from unittest.mock import patch, Mock, mock_open

import pytest

from neo4j_graphrag.utils.file_handler import FileHandler, FileFormat


def test_file_handler_guess_format() -> None:
    handler = FileHandler()

    assert handler._guess_file_format(Path("file.json")) == FileFormat.JSON
    assert handler._guess_file_format(Path("file.JSON")) == FileFormat.JSON
    assert handler._guess_file_format(Path("file.yaml")) == FileFormat.YAML
    assert handler._guess_file_format(Path("file.YAML")) == FileFormat.YAML
    assert handler._guess_file_format(Path("file.yml")) == FileFormat.YAML
    assert handler._guess_file_format(Path("file.YML")) == FileFormat.YAML
    assert handler._guess_file_format(Path("file.txt")) is None


@patch("neo4j_graphrag.utils.file_handler.FileHandler._read_json")
@patch("neo4j_graphrag.utils.file_handler.FileHandler._check_file_exists")
def test_file_handler_read_json_from_read_method_happy_path(
    mock_file_exists: Mock,
    mock_read_json: Mock,
) -> None:
    handler = FileHandler()

    mock_file_exists.return_value = Path("file.json")
    mock_read_json.return_value = {}
    data = handler.read("file.json")
    mock_read_json.assert_called_with(Path("file.json"))
    assert data == {}

    mock_file_exists.return_value = Path("file.JSON")
    mock_read_json.return_value = {}
    data = handler.read("file.JSON")
    mock_read_json.assert_called_with(Path("file.JSON"))
    assert data == {}


@patch("neo4j_graphrag.utils.file_handler.FileHandler._read_yaml")
@patch("neo4j_graphrag.utils.file_handler.FileHandler._check_file_exists")
def test_file_handler_read_yaml_from_read_method_happy_path(
    mock_file_exists: Mock,
    mock_read_yaml: Mock,
) -> None:
    mock_file_exists.return_value = Path("file.yaml")
    handler = FileHandler()
    mock_read_yaml.return_value = {}
    data = handler.read("file.yaml")
    mock_read_yaml.assert_called_with(Path("file.yaml"))
    assert data == {}

    mock_file_exists.return_value = Path("file.yml")
    mock_read_yaml.return_value = {}
    data = handler.read("file.yml")
    mock_read_yaml.assert_called_with(Path("file.yml"))
    assert data == {}

    mock_file_exists.return_value = Path("file.YAML")
    mock_read_yaml.return_value = {}
    data = handler.read("file.YAML")
    mock_read_yaml.assert_called_with(Path("file.YAML"))
    assert data == {}


@patch("neo4j_graphrag.utils.file_handler.LocalFileSystem")
def test_file_handler_read_json_method_happy_path(
    mock_fs: Mock,
) -> None:
    mock_fs_open = mock_open(read_data='{"data": 1}')
    mock_fs.return_value.open = mock_fs_open

    handler = FileHandler()
    data = handler._read_json(Path("file.json"))
    mock_fs_open.assert_called_once_with("file.json", "r")
    assert data == {"data": 1}


@patch("neo4j_graphrag.utils.file_handler.LocalFileSystem")
def test_file_handler_read_yaml_method_happy_path(
    mock_fs: Mock,
) -> None:
    mock_fs_open = mock_open(
        read_data="""
    data: 1
    """
    )
    mock_fs.return_value.open = mock_fs_open

    handler = FileHandler()
    data = handler._read_yaml(Path("file.yaml"))
    mock_fs_open.assert_called_once_with("file.yaml", "r")
    assert data == {"data": 1}


@patch("neo4j_graphrag.utils.file_handler.LocalFileSystem")
def test_file_handler_read_json_file_does_not_exist(
    mock_fs: Mock,
) -> None:
    mock_fs.return_value.open.side_effect = FileNotFoundError
    handler = FileHandler()
    with pytest.raises(FileNotFoundError):
        handler._read_json(Path("file.json"))


@patch("neo4j_graphrag.utils.file_handler.LocalFileSystem")
def test_file_handler_read_json_invalid_json(
    mock_fs: Mock,
) -> None:
    mock_fs_open = mock_open(read_data="{")
    mock_fs.return_value.open = mock_fs_open

    handler = FileHandler()
    with pytest.raises(ValueError, match="Invalid JSON"):
        handler._read_json(Path("file.json"))


@patch("neo4j_graphrag.utils.file_handler.LocalFileSystem")
def test_file_handler_read_yaml_file_does_not_exist(
    mock_fs: Mock,
) -> None:
    mock_fs.return_value.open.side_effect = FileNotFoundError
    handler = FileHandler()
    with pytest.raises(FileNotFoundError):
        handler._read_yaml(Path("file.yaml"))


@patch("neo4j_graphrag.utils.file_handler.LocalFileSystem")
def test_file_handler_read_yaml_invalid_yaml(
    mock_fs: Mock,
) -> None:
    mock_fs_open = mock_open(
        read_data="""
    data: [
    """
    )
    mock_fs.return_value.open = mock_fs_open

    handler = FileHandler()
    with pytest.raises(ValueError, match="Invalid YAML"):
        handler._read_yaml(Path("file.yaml"))


@patch("neo4j_graphrag.utils.file_handler.FileHandler._check_file_exists")
def test_file_handler_file_can_be_written_file(mock_file_exists: Mock) -> None:
    # file does not exist
    mock_file_exists.side_effect = FileNotFoundError()
    handler = FileHandler()
    # nothing happens = all good
    handler._check_file_can_be_written(Path("file.json"))
    # file exist, overwrite is False
    mock_file_exists.side_effect = None
    handler = FileHandler()
    with pytest.raises(ValueError):
        handler._check_file_can_be_written(Path("file.json"), overwrite=False)
    # file exists, overwrite is True
    mock_file_exists.side_effect = None
    handler = FileHandler()
    # nothing happens = all good
    handler._check_file_can_be_written(Path("file.json"), overwrite=True)


@patch("neo4j_graphrag.utils.file_handler.LocalFileSystem")
@patch("neo4j_graphrag.utils.file_handler.json")
def test_file_handler_write_json_happy_path(
    mock_json_module: Mock,
    mock_fs: Mock,
) -> None:
    mock_fs_open = mock_open()
    mock_fs.return_value.open = mock_fs_open
    file_handler = FileHandler()
    file_handler._write_json({"some": "data"}, Path("file.json"))
    mock_fs_open.assert_called_once_with("file.json", "w")
    mock_json_module.dump.assert_called_with(
        {"some": "data"}, mock_fs_open.return_value, indent=2
    )


@patch("neo4j_graphrag.utils.file_handler.LocalFileSystem")
@patch("neo4j_graphrag.utils.file_handler.json")
def test_file_handler_write_json_extra_kwargs_happy_path(
    mock_json_module: Mock,
    mock_fs: Mock,
) -> None:
    mock_fs_open = mock_open()
    mock_fs.return_value.open = mock_fs_open
    file_handler = FileHandler()
    file_handler._write_json({"some": "data"}, Path("file.json"), indent=4, default=str)
    mock_fs_open.assert_called_once_with("file.json", "w")
    mock_json_module.dump.assert_called_with(
        {"some": "data"}, mock_fs_open.return_value, indent=4, default=str
    )


@patch("neo4j_graphrag.utils.file_handler.LocalFileSystem")
@patch("neo4j_graphrag.utils.file_handler.yaml")
def test_file_handler_write_yaml_happy_path(
    mock_yaml_module: Mock,
    mock_fs: Mock,
) -> None:
    mock_fs_open = mock_open()
    mock_fs.return_value.open = mock_fs_open
    file_handler = FileHandler()
    file_handler._write_yaml({"some": "data"}, Path("file.yaml"))
    mock_fs_open.assert_called_once_with("file.yaml", "w")
    mock_yaml_module.safe_dump.assert_called_with(
        {"some": "data"},
        mock_fs_open.return_value,
        default_flow_style=False,
        sort_keys=True,
    )


@patch("neo4j_graphrag.utils.file_handler.LocalFileSystem")
@patch("neo4j_graphrag.utils.file_handler.yaml")
def test_file_handler_write_yaml_extra_kwargs_happy_path(
    mock_yaml_module: Mock,
    mock_fs: Mock,
) -> None:
    mock_fs_open = mock_open()
    mock_fs.return_value.open = mock_fs_open
    file_handler = FileHandler()
    file_handler._write_yaml(
        {"some": "data"}, Path("file.json"), default_flow_style="toto", other_keyword=42
    )
    mock_fs_open.assert_called_once_with("file.json", "w")
    mock_yaml_module.safe_dump.assert_called_with(
        {"some": "data"},
        mock_fs_open.return_value,
        default_flow_style="toto",
        sort_keys=True,
        other_keyword=42,
    )


@patch("neo4j_graphrag.utils.file_handler.FileHandler._write_json")
@patch("neo4j_graphrag.utils.file_handler.FileHandler._check_file_can_be_written")
def test_file_handler_write_json_from_write_method_happy_path(
    mock_file_can_be_written: Mock,
    mock_write_json: Mock,
) -> None:
    handler = FileHandler()
    handler.write("data", "file.json")
    mock_write_json.assert_called_with("data", Path("file.json"))


@patch("neo4j_graphrag.utils.file_handler.FileHandler._write_yaml")
@patch("neo4j_graphrag.utils.file_handler.FileHandler._check_file_can_be_written")
def test_file_handler_write_yaml_from_write_method_happy_path(
    mock_file_can_be_written: Mock,
    mock_write_yaml: Mock,
) -> None:
    handler = FileHandler()
    handler.write("data", "file.yaml")
    mock_write_yaml.assert_called_with("data", Path("file.yaml"))


@patch("neo4j_graphrag.utils.file_handler.FileHandler._write_yaml")
@patch("neo4j_graphrag.utils.file_handler.FileHandler._check_file_can_be_written")
def test_file_handler_write_yaml_from_write_method_overwrite_format_happy_path(
    mock_file_can_be_written: Mock,
    mock_write_yaml: Mock,
) -> None:
    handler = FileHandler()
    handler.write("data", "file.txt", format=FileFormat.YAML)
    mock_write_yaml.assert_called_with("data", Path("file.txt"))


@patch("neo4j_graphrag.utils.file_handler.FileHandler._write_json")
@patch("neo4j_graphrag.utils.file_handler.FileHandler._check_file_can_be_written")
def test_file_handler_write_json_from_write_method_overwrite_format_happy_path(
    mock_file_can_be_written: Mock,
    mock_write_json: Mock,
) -> None:
    handler = FileHandler()
    handler.write("data", "file.txt", format=FileFormat.JSON)
    mock_write_json.assert_called_with("data", Path("file.txt"))



import neo4j
from neo4j_graphrag.indexes import create_vector_index

NEO4J_URL = "neo4j://localhost:7687"
NEO4J_AUTH = ("neo4j", "password")
INDEX_NAME = "vector_index"
DIMENSION = 1536

driver = neo4j.GraphDatabase.driver(NEO4J_URL, auth=NEO4J_AUTH)

create_vector_index(
    driver,
    INDEX_NAME,
    label="Document",
    embedding_property="vectorProperty",
    dimensions=DIMENSION,
    similarity_fn="euclidean",
)



#  Copyright (c) "Neo4j"
#  Neo4j Sweden AB [https://neo4j.com]
#  #
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  #
#      https://www.apache.org/licenses/LICENSE-2.0
#  #
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
from unittest.mock import MagicMock

import neo4j
import pytest
from neo4j_graphrag.experimental.components.entity_relation_extractor import (
    LLMEntityRelationExtractor,
)
from neo4j_graphrag.experimental.components.neo4j_reader import Neo4jChunkReader
from neo4j_graphrag.experimental.components.types import LexicalGraphConfig, TextChunk
from neo4j_graphrag.experimental.pipeline import Pipeline
from neo4j_graphrag.llm import LLMResponse


@pytest.mark.asyncio
@pytest.mark.usefixtures("setup_neo4j_for_kg_construction_with_chunks")
async def test_neo4j_reader(driver: neo4j.Driver) -> None:
    reader = Neo4jChunkReader(driver)
    res = await reader.run()
    assert len(res.chunks) == 2
    assert res.chunks[0] == TextChunk(
        index=0,
        text="some text",
        metadata={"embedding": None},
        uid="0",
    )
    assert res.chunks[1] == TextChunk(
        index=1,
        text="some longer text",
        metadata={"embedding": None},
        uid="1",
    )


@pytest.mark.asyncio
@pytest.mark.usefixtures("setup_neo4j_for_kg_construction_with_chunks")
async def test_neo4j_reader_in_pipeline(driver: neo4j.Driver, llm: MagicMock) -> None:
    llm.ainvoke.side_effect = [
        LLMResponse(
            content="""{
                "nodes": [
                    {
                        "id": "0",
                        "label": "Person",
                        "properties": {
                            "name": "Harry Potter"
                        }
                    }
                ],
                "relationships": []
            }"""
        ),
        LLMResponse(content='{"nodes": [], "relationships": []}'),
    ]
    pipeline = Pipeline()
    pipeline.add_component(Neo4jChunkReader(driver), "reader")
    pipeline.add_component(
        LLMEntityRelationExtractor(llm, create_lexical_graph=False), "extractor"
    )
    pipeline.connect("reader", "extractor", {"chunks": "reader"})
    pipeline_output = await pipeline.run({})
    created_graph = pipeline_output.result["extractor"]
    assert len(created_graph["nodes"]) == 1
    # no lexical graph, so no relationship to the chunk
    assert len(created_graph["relationships"]) == 0


@pytest.mark.asyncio
@pytest.mark.usefixtures("setup_neo4j_for_kg_construction_with_chunks")
async def test_neo4j_reader_in_pipeline_with_lexical_graph_config(
    driver: neo4j.Driver, llm: MagicMock
) -> None:
    llm.ainvoke.side_effect = [
        LLMResponse(
            content="""{
                "nodes": [
                    {
                        "id": "0",
                        "label": "Person",
                        "properties": {
                            "name": "Harry Potter"
                        }
                    }
                ],
                "relationships": []
            }"""
        ),
        LLMResponse(content='{"nodes": [], "relationships": []}'),
    ]
    pipeline = Pipeline()
    pipeline.add_component(Neo4jChunkReader(driver), "reader")
    pipeline.add_component(
        LLMEntityRelationExtractor(llm, create_lexical_graph=False), "extractor"
    )
    pipeline.connect("reader", "extractor", {"chunks": "reader"})
    lg_config = LexicalGraphConfig(node_to_chunk_relationship_type="COMES_FROM")
    pipeline_output = await pipeline.run(
        {
            "extractor": {
                "lexical_graph_config": lg_config,
            }
        }
    )
    created_graph = pipeline_output.result["extractor"]
    assert len(created_graph["nodes"]) == 1
    assert len(created_graph["relationships"]) == 1  # entity to chunk relationship
    assert (
        created_graph["relationships"][0]["type"]
        == lg_config.node_to_chunk_relationship_type
        == "COMES_FROM"
    )



from neo4j_graphrag.experimental.components.text_splitters.base import TextSplitter
from neo4j_graphrag.experimental.components.types import TextChunk, TextChunks


class MySplitter(TextSplitter):
    async def run(self, text: str) -> TextChunks:
        # your logic here
        return TextChunks(
            chunks=[
                TextChunk(text="", index=0),
                # optional metadata
                TextChunk(text="", index=1, metadata={"key": "value"}),
            ]
        )



"""This example uses an example Movie database where movies' plots are embedded
using OpenAI embeddings. OPENAI_API_KEY needs to be set in the environment for
this example to run.

It shows how to use a vector-only retriever to find context
similar to a query **text** using vector similarity.
"""

import neo4j
from neo4j_graphrag.embeddings.openai import OpenAIEmbeddings
from neo4j_graphrag.retrievers import VectorRetriever

# Define database credentials
URI = "neo4j+s://demo.neo4jlabs.com"
AUTH = ("recommendations", "recommendations")
DATABASE = "recommendations"
INDEX_NAME = "moviePlotsEmbedding"


with neo4j.GraphDatabase.driver(URI, auth=AUTH) as driver:
    # Initialize the retriever
    retriever = VectorRetriever(
        driver=driver,
        index_name=INDEX_NAME,
        embedder=OpenAIEmbeddings(),
        # optionally, provide a list of properties to fetch (default fetch all)
        # return_properties=[],
        # optionally, configure how to format the results
        # (see corresponding example in 'customize' directory)
        # result_formatter=None,
        # optionally, set neo4j database
        neo4j_database=DATABASE,
    )

    # Perform the similarity search for a text query
    # (retrieve the top 5 most similar nodes)
    query_text = "Find me a movie about aliens"
    print(retriever.search(query_text=query_text, top_k=5))



#  Copyright (c) "Neo4j"
#  Neo4j Sweden AB [https://neo4j.com]
#  #
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  #
#      https://www.apache.org/licenses/LICENSE-2.0
#  #
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
"""Filename collision handler for Parquet file writing."""

from __future__ import annotations

from pathlib import Path
from typing import Union


class FilenameCollisionHandler:
    """Handles filename collisions by adding numeric suffixes.

    Tracks filename collisions per output path and generates unique filenames
    by appending _n suffixes when the same base filename is requested more
    than once for the same output path.

    Example:

    .. code-block:: python

        handler = FilenameCollisionHandler()
        filename1 = handler.get_unique_filename("Person.parquet", Path("./out"))
        # Returns: "Person.parquet"
        filename2 = handler.get_unique_filename("Person.parquet", Path("./out"))
        # Returns: "Person_1.parquet"
        filename3 = handler.get_unique_filename("Person.parquet", Path("./out"))
        # Returns: "Person_2.parquet"
    """

    # Class-level dictionary to track filename collisions across all instances
    _filename_counts: dict[str, int] = {}

    def get_unique_filename(
        self,
        base_filename: str,
        output_path: Union[str, Path],
    ) -> str:
        """Return a unique filename by adding a _n suffix if a collision is detected.

        Args:
            base_filename: The original filename (e.g. "Person.parquet").
            output_path: The output directory path; collisions are tracked per path.

        Returns:
            A unique filename (e.g. "Person.parquet" or "Person_1.parquet").
        """
        path_str = str(Path(output_path).resolve())
        key = f"{path_str}{base_filename}"

        if key not in self._filename_counts:
            self._filename_counts[key] = 0
            return base_filename

        self._filename_counts[key] += 1
        count = self._filename_counts[key]
        if base_filename.endswith(".parquet"):
            name_without_ext = base_filename[: -len(".parquet")]
            return f"{name_without_ext}_{count}.parquet"
        return f"{base_filename}_{count}"

    @classmethod
    def reset(cls) -> None:
        """Clear the collision-tracking state.

        Intended for tests so each run starts with a clean state.
        """
        cls._filename_counts.clear()



#  Copyright (c) "Neo4j"
#  Neo4j Sweden AB [https://neo4j.com]
#  #
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  #
#      https://www.apache.org/licenses/LICENSE-2.0
#  #
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
from __future__ import annotations

import asyncio
import enum
import json
import logging
from typing import Any, List, Optional, Union

import json_repair
from pydantic import ValidationError, validate_call

from neo4j_graphrag.exceptions import LLMGenerationError
from neo4j_graphrag.experimental.components.lexical_graph import LexicalGraphBuilder
from neo4j_graphrag.experimental.components.schema import GraphSchema
from neo4j_graphrag.experimental.components.types import (
    DocumentInfo,
    LexicalGraphConfig,
    Neo4jGraph,
    TextChunk,
    TextChunks,
)
from neo4j_graphrag.experimental.pipeline.component import Component
from neo4j_graphrag.experimental.pipeline.exceptions import InvalidJSONError
from neo4j_graphrag.generation.prompts import ERExtractionTemplate, PromptTemplate
from neo4j_graphrag.llm import LLMInterface
from neo4j_graphrag.types import LLMMessage
from neo4j_graphrag.utils.logging import prettify

logger = logging.getLogger(__name__)


class OnError(enum.Enum):
    RAISE = "RAISE"
    IGNORE = "IGNORE"

    @classmethod
    def possible_values(cls) -> List[str]:
        return [e.value for e in cls]


def balance_curly_braces(json_string: str) -> str:
    """
    Balances curly braces `{}` in a JSON string. This function ensures that every opening brace has a corresponding
    closing brace, but only when they are not part of a string value. If there are unbalanced closing braces,
    they are ignored. If there are missing closing braces, they are appended at the end of the string.

    Args:
        json_string (str): A potentially malformed JSON string with unbalanced curly braces.

    Returns:
        str: A JSON string with balanced curly braces.
    """
    stack = []
    fixed_json = []
    in_string = False
    escape = False

    for char in json_string:
        if char == '"' and not escape:
            in_string = not in_string
        elif char == "\\" and in_string:
            escape = not escape
            fixed_json.append(char)
            continue
        else:
            escape = False

        if not in_string:
            if char == "{":
                stack.append(char)
                fixed_json.append(char)
            elif char == "}" and stack and stack[-1] == "{":
                stack.pop()
                fixed_json.append(char)
            elif char == "}" and (not stack or stack[-1] != "{"):
                continue
            else:
                fixed_json.append(char)
        else:
            fixed_json.append(char)

    # If stack is not empty, add missing closing braces
    while stack:
        stack.pop()
        fixed_json.append("}")

    return "".join(fixed_json)


def fix_invalid_json(raw_json: str) -> str:
    repaired_json = json_repair.repair_json(raw_json)
    repaired_json = repaired_json.strip()

    if repaired_json == '""':
        raise InvalidJSONError("JSON repair resulted in an empty or invalid JSON.")
    if not repaired_json:
        raise InvalidJSONError("JSON repair resulted in an empty string.")
    return repaired_json


class EntityRelationExtractor(Component):
    """Abstract class for entity relation extraction components.

    Args:
        on_error (OnError): What to do when an error occurs during extraction. Defaults to raising an error.
        create_lexical_graph (bool): Whether to include the text chunks in the graph in addition to the extracted entities and relations. Defaults to True.
    """

    def __init__(
        self,
        *args: Any,
        on_error: OnError = OnError.IGNORE,
        create_lexical_graph: bool = True,
        **kwargs: Any,
    ) -> None:
        self.on_error = on_error
        self.create_lexical_graph = create_lexical_graph

    async def run(
        self,
        chunks: TextChunks,
        document_info: Optional[DocumentInfo] = None,
        lexical_graph_config: Optional[LexicalGraphConfig] = None,
        **kwargs: Any,
    ) -> Neo4jGraph:
        raise NotImplementedError()

    def update_ids(
        self,
        graph: Neo4jGraph,
        chunk: TextChunk,
    ) -> Neo4jGraph:
        """Make node IDs unique across chunks, document and pipeline runs
        by prefixing them with a unique prefix.
        """
        prefix = chunk.chunk_id
        for node in graph.nodes:
            node.id = f"{prefix}:{node.id}"
            if node.properties is None:
                node.properties = {}
        for rel in graph.relationships:
            rel.start_node_id = f"{prefix}:{rel.start_node_id}"
            rel.end_node_id = f"{prefix}:{rel.end_node_id}"
        return graph


class LLMEntityRelationExtractor(EntityRelationExtractor):
    """
    Extracts a knowledge graph from a series of text chunks using a large language model.

    Args:
        llm (LLMInterface): The language model to use for extraction.
        prompt_template (ERExtractionTemplate | str): A custom prompt template to use for extraction.
        create_lexical_graph (bool): Whether to include the text chunks in the graph in addition to the extracted entities and relations. Defaults to True.
        on_error (OnError): What to do when an error occurs during extraction. Defaults to raising an error.
        max_concurrency (int): The maximum number of concurrent tasks which can be used to make requests to the LLM.
        use_structured_output (bool): Whether to use structured output (LLMInterfaceV2) with the Neo4jGraph Pydantic model.
            Only supported for OpenAILLM and VertexAILLM. Defaults to False (uses V1 prompt-based JSON extraction).

    Example with V1 (default, prompt-based JSON):

    .. code-block:: python

        from neo4j_graphrag.experimental.components.entity_relation_extractor import LLMEntityRelationExtractor
        from neo4j_graphrag.llm import OpenAILLM
        from neo4j_graphrag.experimental.pipeline import Pipeline

        llm = OpenAILLM(model_name="gpt-5", model_params={"temperature": 0, "response_format": {"type": "object"}})

        extractor = LLMEntityRelationExtractor(llm=llm)
        pipe = Pipeline()
        pipe.add_component(extractor, "extractor")

    Example with V2 (structured output):

    .. code-block:: python

        from neo4j_graphrag.experimental.components.entity_relation_extractor import LLMEntityRelationExtractor
        from neo4j_graphrag.llm import OpenAILLM
        from neo4j_graphrag.experimental.pipeline import Pipeline

        llm = OpenAILLM(model_name="gpt-5", model_params={"temperature": 0})
        extractor = LLMEntityRelationExtractor(llm=llm, use_structured_output=True)
        pipe = Pipeline()
        pipe.add_component(extractor, "extractor")

    """

    def __init__(
        self,
        llm: LLMInterface,
        prompt_template: Union[ERExtractionTemplate, str] = ERExtractionTemplate(),
        create_lexical_graph: bool = True,
        on_error: OnError = OnError.RAISE,
        max_concurrency: int = 5,
        use_structured_output: bool = False,
    ) -> None:
        super().__init__(on_error=on_error, create_lexical_graph=create_lexical_graph)
        self.llm = llm
        self.max_concurrency = max_concurrency
        self.use_structured_output = use_structured_output

        # Validate that structured output is only used with supported LLMs
        if use_structured_output and not llm.supports_structured_output:
            raise ValueError(
                f"Structured output is not supported by {type(llm).__name__}. "
                f"Please use a model that supports structured output, or set use_structured_output=False."
            )

        if isinstance(prompt_template, str):
            template = PromptTemplate(prompt_template, expected_inputs=[])
        else:
            template = prompt_template
        self.prompt_template = template

    async def extract_for_chunk(
        self, schema: GraphSchema, examples: str, chunk: TextChunk
    ) -> Neo4jGraph:
        """Run entity extraction for a given text chunk."""
        prompt = self.prompt_template.format(
            text=chunk.text,
            schema=schema.model_dump(exclude_none=True),
            examples=examples,
        )

        # Use structured output (V2) if enabled
        if self.use_structured_output:
            # Capability check
            # This should never happen due to __init__ validation
            if not self.llm.supports_structured_output:
                raise RuntimeError(
                    f"Structured output is not supported by {type(self.llm).__name__}"
                )

            messages = [LLMMessage(role="user", content=prompt)]
            llm_result = await self.llm.ainvoke(messages, response_format=Neo4jGraph)  # type: ignore[call-arg, arg-type]
            try:
                chunk_graph = Neo4jGraph.model_validate_json(llm_result.content)
            except ValidationError as e:
                if self.on_error == OnError.RAISE:
                    raise LLMGenerationError("LLM response has improper format") from e
                logger.error(
                    f"LLM response has improper format for chunk_index={chunk.index}"
                )
                logger.debug(f"Invalid response: {llm_result.content}")
                chunk_graph = Neo4jGraph()
            return chunk_graph

        # Use V1 prompt-based JSON extraction (default)
        llm_result = await self.llm.ainvoke(prompt)
        try:
            llm_generated_json = fix_invalid_json(llm_result.content)
            result = json.loads(llm_generated_json)
        except (json.JSONDecodeError, InvalidJSONError) as e:
            if self.on_error == OnError.RAISE:
                raise LLMGenerationError("LLM response is not valid JSON") from e
            logger.error(
                f"LLM response is not valid JSON for chunk_index={chunk.index}"
            )
            logger.debug(f"Invalid JSON: {llm_result.content}")
            result = {"nodes": [], "relationships": []}
        try:
            chunk_graph = Neo4jGraph.model_validate(result)
        except ValidationError as e:
            if self.on_error == OnError.RAISE:
                raise LLMGenerationError("LLM response has improper format") from e
            logger.error(
                f"LLM response has improper format for chunk_index={chunk.index}"
            )
            logger.debug(f"Invalid JSON format: {result}")
            chunk_graph = Neo4jGraph()
        return chunk_graph

    async def post_process_chunk(
        self,
        chunk_graph: Neo4jGraph,
        chunk: TextChunk,
        lexical_graph_builder: Optional[LexicalGraphBuilder] = None,
    ) -> None:
        """Perform post-processing after entity and relation extraction:
        - Update node IDs to make them unique across chunks
        - Build the lexical graph if requested
        """
        self.update_ids(chunk_graph, chunk)
        if lexical_graph_builder:
            await lexical_graph_builder.process_chunk_extracted_entities(
                chunk_graph,
                chunk,
            )

    def combine_chunk_graphs(
        self, lexical_graph: Optional[Neo4jGraph], chunk_graphs: List[Neo4jGraph]
    ) -> Neo4jGraph:
        """Combine sub-graphs obtained for each chunk into a single Neo4jGraph object"""
        if lexical_graph:
            graph = lexical_graph.model_copy(deep=True)
        else:
            graph = Neo4jGraph()
        for chunk_graph in chunk_graphs:
            graph.nodes.extend(chunk_graph.nodes)
            graph.relationships.extend(chunk_graph.relationships)
        return graph

    async def run_for_chunk(
        self,
        sem: asyncio.Semaphore,
        chunk: TextChunk,
        schema: GraphSchema,
        examples: str,
        lexical_graph_builder: Optional[LexicalGraphBuilder] = None,
    ) -> Neo4jGraph:
        """Run extraction, validation and post processing for a single chunk"""
        async with sem:
            chunk_graph = await self.extract_for_chunk(schema, examples, chunk)
            # final_chunk_graph = self.validate_chunk(chunk_graph, schema)
            await self.post_process_chunk(
                chunk_graph,
                chunk,
                lexical_graph_builder,
            )
            return chunk_graph

    @validate_call
    async def run(
        self,
        chunks: TextChunks,
        document_info: Optional[DocumentInfo] = None,
        lexical_graph_config: Optional[LexicalGraphConfig] = None,
        schema: Optional[GraphSchema] = None,
        examples: str = "",
        **kwargs: Any,
    ) -> Neo4jGraph:
        """Perform entity and relation extraction for all chunks in a list.

        Optionally, creates the "lexical graph" by adding nodes and relationships
        to represent the document and its chunks in the returned graph
        (For more details, see the :ref:`Lexical Graph Builder doc <lexical-graph-builder>` and
        the :ref:`User Guide <lexical-graph-in-er-extraction>`)

        Args:
            chunks (TextChunks): List of text chunks to extract entities and relations from.
            document_info (Optional[DocumentInfo], optional): Document the chunks are coming from. Used in the lexical graph creation step.
            lexical_graph_config (Optional[LexicalGraphConfig], optional): Lexical graph configuration to customize node labels and relationship types in the lexical graph.
            schema (GraphSchema | None): Definition of the schema to guide the LLM in its extraction.
            examples (str): Examples for few-shot learning in the prompt.
        """
        lexical_graph_builder = None
        lexical_graph = None
        if self.create_lexical_graph:
            config = lexical_graph_config or LexicalGraphConfig()
            lexical_graph_builder = LexicalGraphBuilder(config=config)
            lexical_graph_result = await lexical_graph_builder.run(
                text_chunks=chunks, document_info=document_info
            )
            lexical_graph = lexical_graph_result.graph
        elif lexical_graph_config:
            lexical_graph_builder = LexicalGraphBuilder(config=lexical_graph_config)
        schema = schema or GraphSchema(
            node_types=(),
        )
        examples = examples or ""
        sem = asyncio.Semaphore(self.max_concurrency)
        tasks = [
            self.run_for_chunk(
                sem,
                chunk,
                schema,
                examples,
                lexical_graph_builder,
            )
            for chunk in chunks.chunks
        ]
        chunk_graphs: list[Neo4jGraph] = list(await asyncio.gather(*tasks))
        graph = self.combine_chunk_graphs(lexical_graph, chunk_graphs)
        logger.debug(f"Extracted graph: {prettify(graph)}")
        return graph



#  Copyright (c) "Neo4j"
#  Neo4j Sweden AB [https://neo4j.com]
#  #
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  #
#      https://www.apache.org/licenses/LICENSE-2.0
#  #
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
from __future__ import annotations

from typing import Any, Optional

from neo4j_graphrag.embeddings.base import Embedder
from neo4j_graphrag.exceptions import EmbeddingsGenerationError
from neo4j_graphrag.utils.rate_limit import RateLimitHandler, rate_limit_handler

try:
    import cohere
except ImportError:
    cohere = None  # type: ignore


class CohereEmbeddings(Embedder):
    def __init__(
        self,
        model: str = "",
        rate_limit_handler: Optional[RateLimitHandler] = None,
        **kwargs: Any,
    ) -> None:
        if cohere is None:
            raise ImportError(
                """Could not import cohere python client.
                Please install it with `pip install "neo4j-graphrag[cohere]"`."""
            )
        super().__init__(rate_limit_handler)
        self.model = model
        self.client = cohere.Client(**kwargs)

    @rate_limit_handler
    def embed_query(self, text: str, **kwargs: Any) -> list[float]:
        try:
            response = self.client.embed(
                texts=[text],
                model=self.model,
                **kwargs,
            )
            return response.embeddings[0]  # type: ignore
        except Exception as e:
            raise EmbeddingsGenerationError(
                f"Failed to generate embedding with Cohere: {e}"
            ) from e



"""Use the PdfLoader component to extract text from a PDF file."""

import asyncio
from pathlib import Path

from neo4j_graphrag.experimental.components.pdf_loader import PdfLoader

root_dir = Path(__file__).parents[4]
file_path = root_dir / "data" / "Harry Potter and the Chamber of Secrets Summary.pdf"


async def main() -> None:
    loader = PdfLoader()
    document = await loader.run(filepath=file_path)
    print(document.text[:200])
    print(document.document_info)


if __name__ == "__main__":
    asyncio.run(main())



"""The SinglePropertyExactMatchResolver merge nodes with same label
and exact same property value (by default using the "name" property).

If some nodes need to be excluded from the resolution, for instance nodes
created from a previous run, a "WHERE" query can be added. The only variable
in the query scope is "entity".

WARNING: this process is destructive, initial nodes are deleted and replaced
by the resolved ones, but all relationships are kept.
See apoc.refactor.mergeNodes documentation for more details.
"""

import neo4j
from neo4j_graphrag.experimental.components.resolver import (
    SinglePropertyExactMatchResolver,
)
from neo4j_graphrag.experimental.components.types import ResolutionStats


async def main(driver: neo4j.Driver) -> None:
    resolver = SinglePropertyExactMatchResolver(
        driver,
        # let's filter out some entities assuming the EntityToExclude label
        # was manually added to nodes in the db
        filter_query="WHERE NOT entity:EntityToExclude",
        # another example: in some cases, we do not want to merge
        # entities whose name is John Doe because we don't know if it
        # corresponds to the same real person
        # filter_query="WHERE entity.name <> 'John Doe'",
        # optionally, change the property used for resolution (default is "name")
        # resolve_property="name",
        # and the neo4j database where data is updated
        # neo4j_database="neo4j",
    )
    res: ResolutionStats = await resolver.run()
    print(res)



#  Copyright (c) "Neo4j"
#  Neo4j Sweden AB [https://neo4j.com]
#  #
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  #
#      https://www.apache.org/licenses/LICENSE-2.0
#  #
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from unittest.mock import MagicMock, call

import neo4j
import pytest
from neo4j_graphrag.exceptions import LLMGenerationError
from neo4j_graphrag.generation.graphrag import GraphRAG
from neo4j_graphrag.generation.types import RagResultModel
from neo4j_graphrag.llm import LLMResponse
from neo4j_graphrag.message_history import Neo4jMessageHistory
from neo4j_graphrag.retrievers import VectorCypherRetriever
from neo4j_graphrag.types import LLMMessage, RetrieverResult, RetrieverResultItem

from tests.e2e.conftest import BiologyEmbedder
from tests.e2e.utils import build_data_objects, populate_neo4j


@pytest.fixture(scope="module")
def populate_neo4j_db(driver: neo4j.Driver) -> None:
    driver.execute_query("MATCH (n) DETACH DELETE n")
    neo4j_objects, q_objects = build_data_objects(q_vector_fmt="neo4j")
    populate_neo4j(driver, neo4j_objects, should_create_vector_index=True)


@pytest.mark.usefixtures("populate_neo4j_db")
def test_graphrag_happy_path(
    driver: MagicMock, llm: MagicMock, biology_embedder: BiologyEmbedder
) -> None:
    retriever = VectorCypherRetriever(
        driver,
        retrieval_query="WITH node RETURN node {.question}",
        index_name="vector-index-name",
        embedder=biology_embedder,
    )
    rag = GraphRAG(
        retriever=retriever,
        llm=llm,
    )
    llm.invoke.return_value = LLMResponse(content="some text")

    result = rag.search(
        query_text="biology",
        retriever_config={
            "top_k": 2,
        },
    )

    llm.invoke.assert_called_once_with(
        input="""Context:
<Record node={'question': 'In 1953 Watson & Crick built a model of the molecular structure of this, the gene-carrying substance'}>
<Record node={'question': 'This organ removes excess glucose from the blood & stores it as glycogen'}>

Examples:


Question:
biology

Answer:
""",
        message_history=None,
        system_instruction="Answer the user question using the provided context.",
    )
    assert isinstance(result, RagResultModel)
    assert result.answer == "some text"
    assert result.retriever_result is None


@pytest.mark.usefixtures("populate_neo4j_db")
def test_graphrag_happy_path_with_neo4j_message_history(
    retriever_mock: MagicMock,
    llm: MagicMock,
    driver: neo4j.Driver,
) -> None:
    rag = GraphRAG(
        retriever=retriever_mock,
        llm=llm,
    )
    retriever_mock.search.return_value = RetrieverResult(
        items=[
            RetrieverResultItem(content="item content 1"),
            RetrieverResultItem(content="item content 2"),
        ]
    )
    llm.invoke.side_effect = [
        LLMResponse(content="llm generated summary"),
        LLMResponse(content="llm generated text"),
    ]
    message_history = Neo4jMessageHistory(
        driver=driver,
        session_id="123",
    )
    message_history.add_messages(
        messages=[
            LLMMessage(role="user", content="initial question"),
            LLMMessage(role="assistant", content="answer to initial question"),
        ]
    )
    res = rag.search(
        query_text="question",
        message_history=message_history,
    )
    expected_retriever_query_text = """
Message Summary:
llm generated summary

Current Query:
question
"""

    first_invocation_input = """
Summarize the message history:

user: initial question
assistant: answer to initial question
"""
    first_invocation_system_instruction = "You are a summarization assistant. Summarize the given text in no more than 300 words."
    second_invocation = """Context:
item content 1
item content 2

Examples:


Question:
question

Answer:
"""
    retriever_mock.search.assert_called_once_with(
        query_text=expected_retriever_query_text
    )
    assert llm.invoke.call_count == 2
    llm.invoke.assert_has_calls(
        [
            call(
                input=first_invocation_input,
                system_instruction=first_invocation_system_instruction,
            ),
            call(
                input=second_invocation,
                message_history=message_history.messages,
                system_instruction="Answer the user question using the provided context.",
            ),
        ]
    )

    assert isinstance(res, RagResultModel)
    assert res.answer == "llm generated text"
    assert res.retriever_result is None
    message_history.clear()


@pytest.mark.usefixtures("populate_neo4j_db")
def test_graphrag_happy_path_return_context(
    driver: MagicMock, llm: MagicMock, biology_embedder: BiologyEmbedder
) -> None:
    retriever = VectorCypherRetriever(
        driver,
        retrieval_query="WITH node RETURN node {.question}",
        index_name="vector-index-name",
        embedder=biology_embedder,
    )
    rag = GraphRAG(
        retriever=retriever,
        llm=llm,
    )
    llm.invoke.return_value = LLMResponse(content="some text")

    result = rag.search(
        query_text="biology",
        retriever_config={
            "top_k": 2,
        },
        return_context=True,
    )

    llm.invoke.assert_called_once_with(
        input="""Context:
<Record node={'question': 'In 1953 Watson & Crick built a model of the molecular structure of this, the gene-carrying substance'}>
<Record node={'question': 'This organ removes excess glucose from the blood & stores it as glycogen'}>

Examples:


Question:
biology

Answer:
""",
        message_history=None,
        system_instruction="Answer the user question using the provided context.",
    )
    assert isinstance(result, RagResultModel)
    assert result.answer == "some text"
    assert isinstance(result.retriever_result, RetrieverResult)
    assert len(result.retriever_result.items) == 2


@pytest.mark.usefixtures("populate_neo4j_db")
def test_graphrag_happy_path_examples(
    driver: MagicMock, llm: MagicMock, biology_embedder: MagicMock
) -> None:
    retriever = VectorCypherRetriever(
        driver,
        retrieval_query="WITH node RETURN node {.question}",
        index_name="vector-index-name",
        embedder=biology_embedder,
    )
    rag = GraphRAG(
        retriever=retriever,
        llm=llm,
    )
    llm.invoke.return_value = LLMResponse(content="some text")

    result = rag.search(
        query_text="biology",
        retriever_config={
            "top_k": 2,
        },
        examples="this is my example",
    )

    llm.invoke.assert_called_once_with(
        input="""Context:
<Record node={'question': 'In 1953 Watson & Crick built a model of the molecular structure of this, the gene-carrying substance'}>
<Record node={'question': 'This organ removes excess glucose from the blood & stores it as glycogen'}>

Examples:
this is my example

Question:
biology

Answer:
""",
        message_history=None,
        system_instruction="Answer the user question using the provided context.",
    )
    assert result.answer == "some text"


@pytest.mark.usefixtures("populate_neo4j_db")
def test_graphrag_llm_error(
    driver: MagicMock, llm: MagicMock, biology_embedder: BiologyEmbedder
) -> None:
    retriever = VectorCypherRetriever(
        driver,
        retrieval_query="WITH node RETURN node {.question}",
        index_name="vector-index-name",
        embedder=biology_embedder,
    )
    rag = GraphRAG(
        retriever=retriever,
        llm=llm,
    )
    llm.invoke.side_effect = LLMGenerationError("error")

    with pytest.raises(LLMGenerationError):
        rag.search(
            query_text="biology",
        )


@pytest.mark.usefixtures("populate_neo4j_db")
def test_graphrag_retrieval_error(
    driver: MagicMock, llm: MagicMock, retriever_mock: MagicMock
) -> None:
    rag = GraphRAG(
        retriever=retriever_mock,
        llm=llm,
    )

    retriever_mock.search.side_effect = TypeError("error")

    with pytest.raises(TypeError):
        rag.search(
            query_text="biology",
        )



#  Copyright (c) "Neo4j"
#  Neo4j Sweden AB [https://neo4j.com]
#  #
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  #
#      https://www.apache.org/licenses/LICENSE-2.0
#  #
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
from typing import Optional

import neo4j


def get_version(
    driver: neo4j.Driver, database: Optional[str] = None
) -> tuple[tuple[int, ...], bool, bool]:
    """
    Retrieves the Neo4j database version and checks if it is running on the Aura platform.

    Args:
        driver (neo4j.Driver): Neo4j Python driver instance to execute the query.
        database (str, optional): The name of the Neo4j database to query. Defaults to None.

    Returns:
        tuple[tuple[int, ...], bool]:
            - A tuple of integers representing the database version (major, minor, patch) or
                (year, month, patch)  for later versions.
            - A boolean indicating whether the database is hosted on the Aura platform.
            - A boolean indicating whether the database is running the enterprise edition.
    """
    records, _, _ = driver.execute_query(
        "CALL dbms.components()",
        database_=database,
        routing_=neo4j.RoutingControl.READ,
    )
    version = records[0]["versions"][0]
    edition = records[0]["edition"]
    # drop everything after the '-' first
    version_main, *_ = version.split("-")
    # convert each number between '.' into int
    version_tuple = tuple(map(int, version_main.split(".")))
    # if no patch version, consider it's 0
    if len(version_tuple) < 3:
        version_tuple = (*version_tuple, 0)
    return version_tuple, "aura" in version, edition == "enterprise"


def is_version_5_23_or_above(version_tuple: tuple[int, ...]) -> bool:
    """
    Determines if the Neo4j database version is 5.23 or above.

    Args:
        version_tuple (tuple[int, ...]): A tuple of integers representing the database version
            (major, minor, patch) or (year, month, patch) for later versions.

    Returns:
        bool: True if the version is 5.23.0 or above, False otherwise.
    """
    return version_tuple >= (5, 23, 0)


def is_version_5_24_or_above(version_tuple: tuple[int, ...]) -> bool:
    """
    Determines if the Neo4j database version is 5.24 or above.

    Dynamic label syntax (SET n:$(labels)) was introduced in Neo4j 5.24,
    replacing the deprecated apoc.create.addLabels procedure.

    Args:
        version_tuple (tuple[int, ...]): A tuple of integers representing the database version
            (major, minor, patch) or (year, month, patch) for later versions.

    Returns:
        bool: True if the version is 5.24.0 or above, False otherwise.
    """
    return version_tuple >= (5, 24, 0)


def has_vector_index_support(version_tuple: tuple[int, ...]) -> bool:
    """
    Checks if a Neo4j database supports vector indexing based on its version and platform.

    Args:
        version_tuple (neo4j.Driver): A tuple of integers representing the database version (major, minor, patch) or
                (year, month, patch)  for later versions.

    Returns:
        bool: True if the connected Neo4j database version supports vector indexing, False otherwise.
    """
    return version_tuple >= (5, 11, 0)


def has_metadata_filtering_support(
    version_tuple: tuple[int, ...], is_aura: bool
) -> bool:
    """
    Checks if a Neo4j database supports vector index metadata filtering based on its version and platform.

    Args:
        version_tuple (neo4j.Driver): A tuple of integers representing the database version (major, minor, patch) or
                (year, month, patch)  for later versions.
        is_aura (bool): A boolean indicating whether the database is hosted on the Aura platform.

    Returns:
        bool: True if the connected Neo4j database version supports vector index metadata filtering , False otherwise.
    """
    if is_aura:
        target_version = (5, 18, 0)
    else:
        target_version = (5, 18, 1)

    return version_tuple >= target_version



#  Copyright (c) "Neo4j"
#  Neo4j Sweden AB [https://neo4j.com]
#  #
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  #
#      https://www.apache.org/licenses/LICENSE-2.0
#  #
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import Callable
from unittest.mock import MagicMock, patch

import neo4j
import pytest
from neo4j_graphrag.embeddings.base import Embedder
from neo4j_graphrag.experimental.pipeline import Component
from neo4j_graphrag.llm import LLMInterface
from neo4j_graphrag.retrievers import (
    HybridRetriever,
    Text2CypherRetriever,
    VectorCypherRetriever,
    VectorRetriever,
)
from neo4j_graphrag.types import RetrieverResultItem


@pytest.fixture(scope="function")
def driver() -> MagicMock:
    return MagicMock(spec=neo4j.Driver)


@pytest.fixture(scope="function")
def embedder() -> MagicMock:
    return MagicMock(spec=Embedder)


@pytest.fixture(scope="function")
def llm() -> MagicMock:
    mock = MagicMock(spec=LLMInterface)
    mock.supports_structured_output = False
    return mock


@pytest.fixture(scope="function")
def retriever_mock() -> MagicMock:
    return MagicMock(spec=VectorRetriever)


@pytest.fixture(scope="function")
@patch("neo4j_graphrag.retrievers.base.get_version")
def vector_retriever(mock_get_version: MagicMock, driver: MagicMock) -> VectorRetriever:
    mock_get_version.return_value = ((5, 23, 0), False, False)
    return VectorRetriever(driver, "my-index")


@pytest.fixture(scope="function")
@patch("neo4j_graphrag.retrievers.base.get_version")
def vector_cypher_retriever(
    mock_get_version: MagicMock, driver: MagicMock
) -> VectorCypherRetriever:
    mock_get_version.return_value = ((5, 23, 0), False, False)
    retrieval_query = "RETURN node.id AS node_id, node.text AS text, score"
    return VectorCypherRetriever(driver, "my-index", retrieval_query)


@pytest.fixture(scope="function")
@patch("neo4j_graphrag.retrievers.base.get_version")
def hybrid_retriever(mock_get_version: MagicMock, driver: MagicMock) -> HybridRetriever:
    mock_get_version.return_value = ((5, 23, 0), False, False)
    return HybridRetriever(driver, "my-index", "my-fulltext-index")


@pytest.fixture(scope="function")
@patch("neo4j_graphrag.retrievers.base.get_version")
def t2c_retriever(
    mock_get_version: MagicMock, driver: MagicMock, llm: MagicMock
) -> Text2CypherRetriever:
    mock_get_version.return_value = ((5, 23, 0), False, False)
    return Text2CypherRetriever(driver, llm)


@pytest.fixture(scope="function")
def neo4j_record() -> neo4j.Record:
    return neo4j.Record({"node": "dummy-node", "score": 1.0, "node_id": 123})


@pytest.fixture(scope="function")
def result_formatter() -> Callable[[neo4j.Record], RetrieverResultItem]:
    def format_function(record: neo4j.Record) -> RetrieverResultItem:
        return RetrieverResultItem(
            content=record.get("node"),
            metadata={"score": record.get("score"), "node_id": record.get("node_id")},
        )

    return format_function


@pytest.fixture(scope="function")
def component() -> MagicMock:
    return MagicMock(spec=Component)



"""This examples shows how to create a custom component
that can be added to a Pipeline with:

c = MyComponent(min_value=0, max_value=10)
pipe = Pipeline()
pipe.add_component(c, name="my_component")
"""

import random

from neo4j_graphrag.experimental.pipeline import Component, DataModel
from pydantic import BaseModel, validate_call


class ComponentInputModel(BaseModel):
    """A class to model the component inputs.
    This is not required, inputs can also be passed individually.

    Note: can also inherit from DataModel.
    """

    text: str


class ComponentResultModel(DataModel):
    """A class to model the component outputs.
    Each component must have such a description of the output,
    so that the parameter mapping can be validated before the
    pipeline run starts.
    """

    value: int
    text: str


class MyComponent(Component):
    """Multiplies an input text by a random number
    between `min_value` and `max_value`
    """

    def __init__(self, min_value: int, max_value: int) -> None:
        self.min_value = min_value
        self.max_value = max_value

    # this decorator is required when a Pydantic model is used in the inputs
    @validate_call
    async def run(self, inputs: ComponentInputModel) -> ComponentResultModel:
        # logic here
        random_value = random.randint(self.min_value, self.max_value)
        return ComponentResultModel(
            value=random_value,
            text=inputs.text * random_value,
        )


if __name__ == "__main__":
    import asyncio

    c = MyComponent(min_value=0, max_value=10)
    print(
        asyncio.run(
            c.run(
                inputs={"text": "Hello"}  # type: ignore
            )
        )
    )



#  Copyright (c) "Neo4j"
#  Neo4j Sweden AB [https://neo4j.com]
#  #
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  #
#      https://www.apache.org/licenses/LICENSE-2.0
#  #
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from __future__ import annotations

from typing import Any, Optional

from neo4j_graphrag.embeddings.base import Embedder
from neo4j_graphrag.exceptions import EmbeddingsGenerationError
from neo4j_graphrag.utils.rate_limit import (
    RateLimitHandler,
    rate_limit_handler,
    async_rate_limit_handler,
)


class OllamaEmbeddings(Embedder):
    """
    Ollama embeddings class.
    This class uses the ollama Python client to generate vector embeddings for text data.

    Args:
        model (str): The name of the Mistral AI text embedding model to use. Defaults to "mistral-embed".
    """

    def __init__(
        self,
        model: str,
        rate_limit_handler: Optional[RateLimitHandler] = None,
        **kwargs: Any,
    ) -> None:
        try:
            import ollama
        except ImportError:
            raise ImportError(
                """Could not import ollama python client.
                Please install it with `pip install "neo4j_graphrag[ollama]"`."""
            )
        super().__init__(rate_limit_handler)
        self.model = model
        self.client = ollama.Client(**kwargs)
        self.async_client = ollama.AsyncClient(**kwargs)

    @rate_limit_handler
    def embed_query(self, text: str, **kwargs: Any) -> list[float]:
        """
        Generate embeddings for a given query using an Ollama text embedding model.

        Args:
            text (str): The text to generate an embedding for.
            **kwargs (Any): Additional keyword arguments to pass to the Ollama client.
        """
        embeddings_response = self.client.embed(
            model=self.model,
            input=text,
            **kwargs,
        )

        if embeddings_response is None or not embeddings_response.embeddings:
            raise EmbeddingsGenerationError("Failed to retrieve embeddings.")

        embeddings = embeddings_response.embeddings
        # client always returns a sequence of sequences
        embedding = embeddings[0]
        if not isinstance(embedding, list):
            raise EmbeddingsGenerationError("Embedding is not a list of floats.")

        return embedding

    @async_rate_limit_handler
    async def async_embed_query(self, text: str, **kwargs: Any) -> list[float]:
        """
        Asynchronously generate embeddings for a given query using an Ollama text embedding model.

        Args:
            text (str): The text to generate an embedding for.
            **kwargs (Any): Additional keyword arguments to pass to the Ollama client.
        """
        embeddings_response = await self.async_client.embed(
            model=self.model,
            input=text,
            **kwargs,
        )

        if embeddings_response is None or not embeddings_response.embeddings:
            raise EmbeddingsGenerationError("Failed to retrieve embeddings.")

        embeddings = embeddings_response.embeddings
        # client always returns a sequence of sequences
        embedding = embeddings[0]
        if not isinstance(embedding, list):
            raise EmbeddingsGenerationError("Embedding is not a list of floats.")

        return embedding



"""This example demonstrate how to invoke an LLM using a local model
served by Ollama.
"""

from neo4j_graphrag.llm import LLMResponse, OllamaLLM

llm = OllamaLLM(
    model_name="<model_name>",
    # model_params={"options": {"temperature": 0}, "format": "json"},
    # host="...",  # if using a remote server
)
res: LLMResponse = llm.invoke("What is the additive color model?")
print(res.content)



from neo4j_graphrag.embeddings import CohereEmbeddings

# set api key here on in the CO_API_KEY env var
api_key = None

embeder = CohereEmbeddings(
    model="embed-english-v3.0",
    api_key=api_key,
)
res = embeder.embed_query("my question")
print(res[:10])



#  Copyright (c) "Neo4j"
#  Neo4j Sweden AB [https://neo4j.com]
#  #
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  #
#      https://www.apache.org/licenses/LICENSE-2.0
#  #
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
from __future__ import annotations

import uuid

import pytest
from neo4j_graphrag.experimental.components.lexical_graph import LexicalGraphBuilder
from neo4j_graphrag.experimental.components.types import (
    DEFAULT_CHUNK_NODE_LABEL,
    DEFAULT_CHUNK_TO_DOCUMENT_RELATIONSHIP_TYPE,
    DEFAULT_DOCUMENT_NODE_LABEL,
    DEFAULT_NEXT_CHUNK_RELATIONSHIP_TYPE,
    DocumentInfo,
    GraphResult,
    LexicalGraphConfig,
    Neo4jNode,
    TextChunk,
    TextChunks,
)


def test_lexical_graph_builder_create_chunk_node_no_metadata() -> None:
    builder = LexicalGraphBuilder()
    node = builder.create_chunk_node(chunk=TextChunk(text="text chunk", index=0))
    assert isinstance(node, Neo4jNode)
    assert node.id is not None
    assert node.properties == {"index": 0, "text": "text chunk"}
    assert node.embedding_properties == {}


def test_lexical_graph_builder_create_chunk_node_metadata_no_embedding() -> None:
    builder = LexicalGraphBuilder()
    node = builder.create_chunk_node(
        chunk=TextChunk(text="text chunk", index=0, metadata={"status": "ok"})
    )
    assert isinstance(node, Neo4jNode)
    assert node.id is not None
    assert node.properties == {"index": 0, "text": "text chunk", "status": "ok"}
    assert node.embedding_properties == {}


def test_lexical_graph_builder_create_chunk_node_metadata_embedding() -> None:
    builder = LexicalGraphBuilder()
    node = builder.create_chunk_node(
        chunk=TextChunk(
            text="text chunk",
            index=0,
            metadata={"status": "ok", "embedding": [1, 2, 3]},
        ),
    )
    assert isinstance(node, Neo4jNode)
    assert node.id is not None
    assert node.properties == {"index": 0, "text": "text chunk", "status": "ok"}
    assert node.embedding_properties == {"embedding": [1, 2, 3]}


@pytest.mark.asyncio
async def test_lexical_graph_builder_run_with_document() -> None:
    lexical_graph_builder = LexicalGraphBuilder()
    doc_uid = str(uuid.uuid4())
    result = await lexical_graph_builder.run(
        text_chunks=TextChunks(
            chunks=[
                TextChunk(text="text chunk 1", index=0),
                TextChunk(text="text chunk 1", index=1),
            ]
        ),
        document_info=DocumentInfo(
            path="test_lexical_graph",
            uid=doc_uid,
            document_type="my_type",
        ),
    )
    assert isinstance(result, GraphResult)
    graph = result.graph
    nodes = graph.nodes
    assert len(nodes) == 3
    document = nodes[0]
    assert document.id == doc_uid
    assert document.label == DEFAULT_DOCUMENT_NODE_LABEL
    assert document.properties["path"] == "test_lexical_graph"
    assert document.properties["createdAt"] is not None
    assert document.properties["document_type"] == "my_type"
    chunk1 = nodes[1]
    assert chunk1.label == DEFAULT_CHUNK_NODE_LABEL
    chunk2 = nodes[2]
    assert chunk2.label == DEFAULT_CHUNK_NODE_LABEL
    assert len(graph.relationships) == 3
    assert graph.relationships[0].type == DEFAULT_CHUNK_TO_DOCUMENT_RELATIONSHIP_TYPE
    assert graph.relationships[1].type == DEFAULT_NEXT_CHUNK_RELATIONSHIP_TYPE
    assert graph.relationships[2].type == DEFAULT_CHUNK_TO_DOCUMENT_RELATIONSHIP_TYPE


@pytest.mark.asyncio
async def test_lexical_graph_builder_run_no_document() -> None:
    lexical_graph_builder = LexicalGraphBuilder()
    result = await lexical_graph_builder.run(
        text_chunks=TextChunks(
            chunks=[
                TextChunk(text="text chunk 1", index=0),
                TextChunk(text="text chunk 1", index=1),
            ]
        ),
    )
    assert isinstance(result, GraphResult)
    graph = result.graph
    nodes = graph.nodes
    assert len(nodes) == 2
    chunk1 = nodes[0]
    assert chunk1.label == DEFAULT_CHUNK_NODE_LABEL
    chunk2 = nodes[1]
    assert chunk2.label == DEFAULT_CHUNK_NODE_LABEL
    assert len(graph.relationships) == 1
    assert graph.relationships[0].type == DEFAULT_NEXT_CHUNK_RELATIONSHIP_TYPE


@pytest.mark.asyncio
async def test_lexical_graph_builder_run_custom_labels() -> None:
    lexical_graph_builder = LexicalGraphBuilder(
        config=LexicalGraphConfig(
            document_node_label="Report",
            chunk_node_label="Page",
            chunk_to_document_relationship_type="IN_REPORT",
            next_chunk_relationship_type="NEXT_PAGE",
        ),
    )
    doc_uid = str(uuid.uuid4())
    result = await lexical_graph_builder.run(
        text_chunks=TextChunks(
            chunks=[
                TextChunk(text="text chunk 1", index=0),
                TextChunk(text="text chunk 1", index=1),
            ]
        ),
        document_info=DocumentInfo(path="test_lexical_graph", uid=doc_uid),
    )
    assert isinstance(result, GraphResult)
    graph = result.graph
    nodes = graph.nodes
    assert len(nodes) == 3
    document = nodes[0]
    assert document.id == doc_uid
    assert document.label == "Report"
    assert document.properties["path"] == "test_lexical_graph"
    chunk1 = nodes[1]
    assert chunk1.label == "Page"
    chunk2 = nodes[2]
    assert chunk2.label == "Page"
    assert len(graph.relationships) == 3
    assert graph.relationships[0].type == "IN_REPORT"
    assert graph.relationships[1].type == "NEXT_PAGE"
    assert graph.relationships[2].type == "IN_REPORT"



#  Copyright (c) "Neo4j"
#  Neo4j Sweden AB [https://neo4j.com]
#  #
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  #
#      https://www.apache.org/licenses/LICENSE-2.0
#  #
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
from __future__ import annotations  # Reminder: May be removed after Python 3.9 is EOL.

import inspect
from typing import Any, Optional
from unittest.mock import MagicMock, patch

import pytest
from neo4j_graphrag.exceptions import Neo4jVersionError
from neo4j_graphrag.retrievers.base import Retriever
from neo4j_graphrag.types import RawSearchResult, RetrieverResult


@pytest.mark.parametrize(
    "db_version,expected_exception",
    [
        (((5, 18, 0), True, True), None),
        (((5, 3, 0), True, True), Neo4jVersionError),
        (((5, 19, 0), False, True), None),
        (((4, 3, 5), False, True), Neo4jVersionError),
        (((5, 23, 0), False, True), None),
        (((2025, 1, 0), False, True), None),
        (((2025, 1, 0), True, True), None),
    ],
)
@patch("neo4j_graphrag.retrievers.base.get_version")
def test_retriever_version_support(
    mock_get_version: MagicMock,
    driver: MagicMock,
    db_version: tuple[tuple[int, ...], bool],
    expected_exception: Optional[type[ValueError]],
) -> None:
    mock_get_version.return_value = db_version

    class MockRetriever(Retriever):
        def get_search_results(self, *args: Any, **kwargs: Any) -> RawSearchResult:
            return RawSearchResult(records=[])

    if expected_exception:
        with pytest.raises(expected_exception):
            MockRetriever(driver=driver)
    else:
        MockRetriever(driver=driver)


@patch("neo4j_graphrag.retrievers.base.get_version")
def test_retriever_search_docstring_copied(
    mock_get_version: MagicMock,
    driver: MagicMock,
) -> None:
    mock_get_version.return_value = ((5, 23, 0), False, False)

    class MockRetriever(Retriever):
        def get_search_results(self, query: str, top_k: int = 10) -> RawSearchResult:
            """My fabulous docstring"""
            return RawSearchResult(records=[])

    retriever = MockRetriever(driver=driver)
    assert retriever.search.__doc__ == "My fabulous docstring"
    signature = inspect.signature(retriever.search)
    assert "query" in signature.parameters
    query_param = signature.parameters["query"]
    assert query_param.default == query_param.empty
    assert query_param.annotation == "str"
    assert "top_k" in signature.parameters
    top_k_param = signature.parameters["top_k"]
    assert top_k_param.default == 10
    assert top_k_param.annotation == "int"


@patch("neo4j_graphrag.retrievers.base.get_version")
def test_retriever_search_docstring_unchanged(
    mock_get_version: MagicMock,
    driver: MagicMock,
) -> None:
    mock_get_version.return_value = ((5, 23, 0), False, False)

    class MockRetrieverForNoise(Retriever):
        def get_search_results(self, query: str, top_k: int = 10) -> RawSearchResult:
            """My fabulous docstring"""
            return RawSearchResult(records=[])

    class MockRetriever(Retriever):
        def get_search_results(self, *args: Any, **kwargs: Any) -> RawSearchResult:
            return RawSearchResult(records=[])

        def search(self, query: str, top_k: int = 10) -> RetrieverResult:
            """My fabulous docstring that I do not want to be updated"""
            return RetrieverResult(items=[])

    assert MockRetrieverForNoise.search is not MockRetriever.search

    retriever = MockRetriever(driver=driver)
    assert (
        retriever.search.__doc__
        == "My fabulous docstring that I do not want to be updated"
    )



import pytest
from neo4j_graphrag.exceptions import (
    PromptMissingInputError,
    PromptMissingPlaceholderError,
)
from neo4j_graphrag.generation.prompts import PromptTemplate


def test_prompt_template_all_default() -> None:
    class MyTemplate(PromptTemplate):
        DEFAULT_TEMPLATE = "My question is {query_text}"
        EXPECTED_INPUTS = ["query_text"]

    template = MyTemplate()
    assert template.template == MyTemplate.DEFAULT_TEMPLATE
    assert template.expected_inputs == MyTemplate.EXPECTED_INPUTS
    assert template.format(query_text="query_text") == "My question is query_text"


def test_prompt_template_overwrite_defaults() -> None:
    class MyTemplate(PromptTemplate):
        DEFAULT_TEMPLATE = "My question is {query_text}"
        EXPECTED_INPUTS = ["query_text"]

    template = MyTemplate(
        template="Please answer my question {query_text} as a {speaker_type}",
        expected_inputs=["query_text", "speaker_type"],
    )
    assert (
        template.template
        == "Please answer my question {query_text} as a {speaker_type}"
    )
    assert template.expected_inputs == ["query_text", "speaker_type"]
    assert (
        template.format(query_text="query_text", speaker_type="child")
        == "Please answer my question query_text as a child"
    )


def test_prompt_template_format_missing_value() -> None:
    class MyTemplate(PromptTemplate):
        EXPECTED_INPUTS = ["query_text", "other"]

    template = MyTemplate(template="{query_text} {other}")
    with pytest.raises(PromptMissingInputError) as excinfo:
        template.format(query_text="query_text")
    assert "Missing input 'other'" in str(excinfo)


def test_prompt_template_format_extra_values() -> None:
    class MyTemplate(PromptTemplate):
        DEFAULT_TEMPLATE = "My question is {query_text} {other}"
        EXPECTED_INPUTS = ["query_text"]

    template = MyTemplate()
    with pytest.raises(KeyError) as excinfo:
        template.format(query_text="query_text")
    assert "KeyError('other')" in str(excinfo)


def test_prompt_template_missing_placeholders() -> None:
    class MyTemplate(PromptTemplate):
        DEFAULT_TEMPLATE = "My question is {query_text} {other}"
        EXPECTED_INPUTS = ["query_text", "banana"]

    with pytest.raises(PromptMissingPlaceholderError) as e:
        MyTemplate()

    assert "`template` is missing placeholder banana" in str(e)


def test_prompt_template_format_given_unused_kwargs() -> None:
    class MyTemplate(PromptTemplate):
        DEFAULT_TEMPLATE = "My question is {query_text}"
        EXPECTED_INPUTS = ["query_text"]

    template = MyTemplate()

    assert (
        template.format(query_text="what do?", banana="b") == "My question is what do?"
    )



#  Copyright (c) "Neo4j"
#  Neo4j Sweden AB [https://neo4j.com]
#  #
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  #
#      https://www.apache.org/licenses/LICENSE-2.0
#  #
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import os
from typing import Any
from unittest.mock import AsyncMock, Mock, patch

import pytest
from neo4j import Driver
from neo4j_graphrag.experimental.pipeline.config.runner import PipelineRunner
from neo4j_graphrag.experimental.pipeline.pipeline import PipelineResult
from neo4j_graphrag.llm import LLMResponse


@pytest.fixture(scope="function", autouse=True)
def clear_db(driver: Driver) -> Any:
    driver.execute_query("MATCH (n) DETACH DELETE n")
    yield


@pytest.mark.asyncio
async def test_pipeline_from_json_config(harry_potter_text: str, driver: Mock) -> None:
    os.environ["NEO4J_URI"] = "neo4j://localhost:7687"
    os.environ["NEO4J_USER"] = "neo4j"
    os.environ["NEO4J_PASSWORD"] = "password"

    runner = PipelineRunner.from_config_file(
        "tests/e2e/data/config_files/pipeline_config.json"
    )
    res = await runner.run({"splitter": {"text": harry_potter_text}})
    assert isinstance(res, PipelineResult)
    meta = res.result["writer"]["metadata"]
    assert "statistics" in meta
    assert meta["statistics"]["node_count"] == 11
    assert meta["statistics"]["relationship_count"] == 10
    assert "nodes_per_label" in meta["statistics"]
    assert "rel_per_type" in meta["statistics"]
    assert "input_files_count" in meta["statistics"]
    assert "input_files_total_size_bytes" in meta["statistics"]
    nodes = driver.execute_query("MATCH (n) RETURN n")
    assert len(nodes.records) == 11


@pytest.mark.asyncio
async def test_pipeline_from_yaml_config(harry_potter_text: str, driver: Mock) -> None:
    os.environ["NEO4J_URI"] = "neo4j://localhost:7687"
    os.environ["NEO4J_USER"] = "neo4j"
    os.environ["NEO4J_PASSWORD"] = "password"

    runner = PipelineRunner.from_config_file(
        "tests/e2e/data/config_files/pipeline_config.yaml"
    )
    res = await runner.run({"splitter": {"text": harry_potter_text}})
    assert isinstance(res, PipelineResult)
    meta = res.result["writer"]["metadata"]
    assert "statistics" in meta
    assert meta["statistics"]["node_count"] == 11
    assert meta["statistics"]["relationship_count"] == 10
    assert "nodes_per_label" in meta["statistics"]
    assert "rel_per_type" in meta["statistics"]
    assert "input_files_count" in meta["statistics"]
    assert "input_files_total_size_bytes" in meta["statistics"]

    nodes = driver.execute_query("MATCH (n) RETURN n")
    assert len(nodes.records) == 11


@patch(
    "neo4j_graphrag.experimental.pipeline.config.runner.SimpleKGPipelineConfig.get_default_embedder"
)
@patch(
    "neo4j_graphrag.experimental.pipeline.config.runner.SimpleKGPipelineConfig.get_default_llm"
)
@pytest.mark.asyncio
async def test_simple_kg_pipeline_from_json_config(
    mock_llm: Mock, mock_embedder: Mock, harry_potter_text: str, driver: Mock
) -> None:
    mock_llm.return_value.ainvoke = AsyncMock(
        side_effect=[
            LLMResponse(
                content="""{
                            "nodes": [
                                {
                                    "id": "0",
                                    "label": "Person",
                                    "properties": {
                                        "name": "Harry Potter"
                                    }
                                },
                                {
                                    "id": "1",
                                    "label": "Person",
                                    "properties": {
                                        "name": "Alastor Mad-Eye Moody"
                                    }
                                },
                                {
                                    "id": "2",
                                    "label": "Organization",
                                    "properties": {
                                        "name": "The Order of the Phoenix"
                                    }
                                }
                            ],
                            "relationships": [
                                {
                                    "type": "KNOWS",
                                    "start_node_id": "0",
                                    "end_node_id": "1"
                                },
                                {
                                    "type": "LED_BY",
                                    "start_node_id": "2",
                                    "end_node_id": "1"
                                }
                            ]
                        }"""
            ),
        ]
    )
    mock_embedder.return_value.async_embed_query = AsyncMock(
        side_effect=[
            [1.0, 2.0],
        ]
    )

    os.environ["NEO4J_URI"] = "neo4j://localhost:7687"
    os.environ["NEO4J_USER"] = "neo4j"
    os.environ["NEO4J_PASSWORD"] = "password"
    os.environ["OPENAI_API_KEY"] = "sk-my-secret-key"
    os.environ["MY_OPENAI_KEY"] = "my-openai-key"

    runner = PipelineRunner.from_config_file(
        "tests/e2e/data/config_files/simple_kg_pipeline_config.json"
    )

    # check extras and API keys are handled as expected
    config = runner.config
    assert config is not None
    # extras must be resolved:
    assert config._global_data["extras"] == {"openai_api_key": "my-openai-key"}
    # API key for LLM is read from env vars (see config file)
    default_llm = config._global_data["llm_config"]["default"]
    assert default_llm.client.api_key == "sk-my-secret-key"
    # API key for embedder is read from extras (see config file)
    default_embedder = config._global_data["embedder_config"]["default"]
    assert default_embedder.client.api_key == "my-openai-key"

    # then run pipeline and check results
    res = await runner.run({"file_path": "tests/e2e/data/documents/harry_potter.pdf"})
    assert isinstance(res, PipelineResult)
    assert res.result["resolver"] == {
        "number_of_nodes_to_resolve": 3,
        "number_of_created_nodes": 3,
    }
    nodes = driver.execute_query("MATCH (n) RETURN n")
    # 1 chunk + 1 document + 3 __Entity__ nodes
    assert len(nodes.records) == 5


@patch(
    "neo4j_graphrag.experimental.pipeline.config.runner.SimpleKGPipelineConfig.get_default_embedder"
)
@patch(
    "neo4j_graphrag.experimental.pipeline.config.runner.SimpleKGPipelineConfig.get_default_llm"
)
@pytest.mark.asyncio
async def test_simple_kg_pipeline_from_yaml_config(
    mock_llm: Mock, mock_embedder: Mock, harry_potter_text: str, driver: Mock
) -> None:
    mock_llm.return_value.ainvoke = AsyncMock(
        side_effect=[
            LLMResponse(
                content="""{
                            "nodes": [
                                {
                                    "id": "0",
                                    "label": "Person",
                                    "properties": {
                                        "name": "Harry Potter"
                                    }
                                },
                                {
                                    "id": "1",
                                    "label": "Person",
                                    "properties": {
                                        "name": "Alastor Mad-Eye Moody"
                                    }
                                },
                                {
                                    "id": "2",
                                    "label": "Organization",
                                    "properties": {
                                        "name": "The Order of the Phoenix"
                                    }
                                }
                            ],
                            "relationships": [
                                {
                                    "type": "KNOWS",
                                    "start_node_id": "0",
                                    "end_node_id": "1"
                                },
                                {
                                    "type": "LED_BY",
                                    "start_node_id": "2",
                                    "end_node_id": "1"
                                }
                            ]
                        }"""
            ),
        ]
    )
    mock_embedder.return_value.async_embed_query = AsyncMock(
        side_effect=[
            [1.0, 2.0],
        ]
    )

    os.environ["NEO4J_URI"] = "neo4j://localhost:7687"
    os.environ["NEO4J_USER"] = "neo4j"
    os.environ["NEO4J_PASSWORD"] = "password"
    os.environ["OPENAI_API_KEY"] = "sk-my-secret-key"

    runner = PipelineRunner.from_config_file(
        "tests/e2e/data/config_files/simple_kg_pipeline_config.yaml"
    )
    res = await runner.run({"file_path": "tests/e2e/data/documents/harry_potter.pdf"})
    assert isinstance(res, PipelineResult)
    # print(await runner.pipeline.store.get_result_for_component(res.run_id, "splitter"))
    assert res.result["resolver"] == {
        "number_of_nodes_to_resolve": 3,
        "number_of_created_nodes": 3,
    }
    nodes = driver.execute_query("MATCH (n) RETURN n")
    # 1 chunk + 1 document + 3 nodes
    assert len(nodes.records) == 5



#  Copyright (c) "Neo4j"
#  Neo4j Sweden AB [https://neo4j.com]
#  #
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  #
#      https://www.apache.org/licenses/LICENSE-2.0
#  #
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
from itertools import zip_longest

import pytest
from neo4j_graphrag.experimental.components.text_splitters.fixed_size_splitter import (
    FixedSizeSplitter,
    _adjust_chunk_end,
    _adjust_chunk_start,
)
from neo4j_graphrag.experimental.components.types import TextChunk


@pytest.mark.asyncio
async def test_split_text_no_overlap() -> None:
    text = "may thy knife chip and shatter"
    chunk_size = 5
    chunk_overlap = 0
    approximate = False
    splitter = FixedSizeSplitter(chunk_size, chunk_overlap, approximate)
    chunks = await splitter.run(text)
    expected_chunks = [
        TextChunk(text="may t", index=0),
        TextChunk(text="hy kn", index=1),
        TextChunk(text="ife c", index=2),
        TextChunk(text="hip a", index=3),
        TextChunk(text="nd sh", index=4),
        TextChunk(text="atter", index=5),
    ]
    for actual, expected in zip_longest(chunks.chunks, expected_chunks):
        assert actual.text == expected.text
        assert actual.index == expected.index
        assert expected.uid is not None


@pytest.mark.asyncio
async def test_split_text_with_overlap() -> None:
    text = "may thy knife chip and shatter"
    chunk_size = 10
    chunk_overlap = 2
    approximate = False
    splitter = FixedSizeSplitter(chunk_size, chunk_overlap, approximate)
    chunks = await splitter.run(text)
    expected_chunks = [
        TextChunk(text="may thy kn", index=0),
        TextChunk(text="knife chip", index=1),
        TextChunk(text="ip and sha", index=2),
        TextChunk(text="hatter", index=3),
    ]
    for actual, expected in zip_longest(chunks.chunks, expected_chunks):
        assert actual.text == expected.text
        assert actual.index == expected.index
        assert expected.uid is not None


@pytest.mark.asyncio
async def test_split_text_empty_string() -> None:
    text = ""
    chunk_size = 5
    chunk_overlap = 1
    approximate = False
    splitter = FixedSizeSplitter(chunk_size, chunk_overlap, approximate)
    chunks = await splitter.run(text)
    assert chunks.chunks == []


def test_invalid_chunk_overlap() -> None:
    with pytest.raises(ValueError) as excinfo:
        FixedSizeSplitter(5, 5)
    assert "chunk_overlap must be strictly less than chunk_size" in str(excinfo)


def test_invalid_chunk_size() -> None:
    with pytest.raises(ValueError) as excinfo:
        FixedSizeSplitter(0, 0)
    assert "chunk_size must be strictly greater than 0" in str(excinfo)


@pytest.mark.parametrize(
    "text, approximate_start, expected_start",
    [
        # Case: approximate_start is at word boundary already
        ("Hello World", 6, 6),
        # Case: approximate_start is at a whitespace already
        ("Hello World", 5, 5),
        # Case: approximate_start is at the middle of word and no whitespace is found
        ("Hello World", 2, 2),
        # Case: approximate_start is at the middle of a word
        ("Hello World", 8, 6),
        # Case: approximate_start = 0
        ("Hello World", 0, 0),
    ],
)
def test_adjust_chunk_start(
    text: str, approximate_start: int, expected_start: int
) -> None:
    """
    Test that the _adjust_chunk_start function correctly shifts
    the start index to avoid breaking words, unless no whitespace is found.
    """
    result = _adjust_chunk_start(text, approximate_start)
    assert result == expected_start


@pytest.mark.parametrize(
    "text, start, approximate_end, expected_end",
    [
        # Case: approximate_end is at word boundary already
        ("Hello World", 0, 5, 5),
        # Case: approximate_end is at the middle of a word
        ("Hello World", 0, 8, 6),
        # Case: approximate_end is at the middle of word and no whitespace is found
        ("Hello World", 0, 3, 3),
        # Case: adjusted_end == start => fallback to approximate_end
        ("Hello World", 6, 7, 7),
        # Case: end>=len(text)
        ("Hello World", 6, 15, 15),
    ],
)
def test_adjust_chunk_end(
    text: str, start: int, approximate_end: int, expected_end: int
) -> None:
    """
    Test that the _adjust_chunk_end function correctly shifts
    the end index to avoid breaking words, unless no whitespace is found.
    """
    result = _adjust_chunk_end(text, start, approximate_end)
    assert result == expected_end


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "text, chunk_size, chunk_overlap, approximate, expected_chunks",
    [
        # Case: approximate fixed size splitting
        (
            "Hello World, this is a test message.",
            10,
            2,
            True,
            ["Hello ", "World, ", "this is a ", "a test ", "message."],
        ),
        # Case: fixed size splitting
        (
            "Hello World, this is a test message.",
            10,
            2,
            False,
            ["Hello Worl", "rld, this ", "s is a tes", "est messag", "age."],
        ),
        # Case: short text => only one chunk
        (
            "Short text",
            20,
            5,
            True,
            ["Short text"],
        ),
        # Case: short text => only one chunk
        (
            "Short text",
            12,
            4,
            True,
            ["Short text"],
        ),
        # Case: text with no spaces
        (
            "1234567890",
            5,
            1,
            True,
            ["12345", "56789", "90"],
        ),
    ],
)
async def test_fixed_size_splitter_run(
    text: str,
    chunk_size: int,
    chunk_overlap: int,
    approximate: bool,
    expected_chunks: list[str],
) -> None:
    """
    Test that 'FixedSizeSplitter.run' returns the expected chunks
    for different configurations.
    """
    splitter = FixedSizeSplitter(
        chunk_size=chunk_size,
        chunk_overlap=chunk_overlap,
        approximate=approximate,
    )
    text_chunks = await splitter.run(text)

    # Verify number of chunks
    assert len(text_chunks.chunks) == len(expected_chunks)

    # Verify content of each chunk
    for i, expected_text in enumerate(expected_chunks):
        assert text_chunks.chunks[i].text == expected_text
        assert isinstance(text_chunks.chunks[i], TextChunk)
        assert text_chunks.chunks[i].index == i



#  Copyright (c) "Neo4j"
#  Neo4j Sweden AB [https://neo4j.com]
#  #
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  #
#      https://www.apache.org/licenses/LICENSE-2.0
#  #
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

# built-in dependencies
from __future__ import annotations

from typing import (
    TYPE_CHECKING,
    Any,
    Iterable,
    List,
    Optional,
    Type,
    Union,
    cast,
    overload,
)

# 3rd party dependencies
from pydantic import BaseModel, ValidationError

# project dependencies
from neo4j_graphrag.exceptions import LLMGenerationError
from neo4j_graphrag.llm.base import LLMInterface, LLMInterfaceV2
from neo4j_graphrag.llm.types import (
    BaseMessage,
    LLMResponse,
    MessageList,
    SystemMessage,
    UserMessage,
)
from neo4j_graphrag.message_history import MessageHistory
from neo4j_graphrag.types import LLMMessage
from neo4j_graphrag.utils.rate_limit import (
    RateLimitHandler,
)
from neo4j_graphrag.utils.rate_limit import (
    async_rate_limit_handler as async_rate_limit_handler_decorator,
)
from neo4j_graphrag.utils.rate_limit import (
    rate_limit_handler as rate_limit_handler_decorator,
)

if TYPE_CHECKING:
    from cohere import ChatMessages


# pylint: disable=redefined-builtin, arguments-differ, raise-missing-from, no-else-return, import-outside-toplevel
class CohereLLM(LLMInterface, LLMInterfaceV2):
    """Interface for large language models on the Cohere platform

    Args:
        model_name (str, optional): Name of the LLM to use. Defaults to "gemini-1.5-flash-001".
        model_params (Optional[dict], optional): Additional parameters for LLMInterface(V1) passed to the model when text is sent to it. Defaults to None.
        system_instruction (Optional[str], optional): Additional instructions for setting the behavior and context for the model in a conversation. Defaults to None.
        rate_limit_handler (Optional[RateLimitHandler], optional): A rate limit handler for LLMInterface(V1) to manage API rate limits. Defaults to None.
        **kwargs (Any): Arguments passed to the model when for the class is initialised. Defaults to None.

    Raises:
        LLMGenerationError: If there's an error generating the response from the model.

    Example:

    .. code-block:: python

        from neo4j_graphrag.llm import CohereLLM

        llm = CohereLLM(api_key="...")
        llm.invoke("Say something")
    """

    def __init__(
        self,
        model_name: str = "",
        model_params: Optional[dict[str, Any]] = None,
        rate_limit_handler: Optional[RateLimitHandler] = None,
        **kwargs: Any,
    ) -> None:
        try:
            import cohere
        except ImportError:
            raise ImportError(
                """Could not import cohere python client.
                Please install it with `pip install "neo4j-graphrag[cohere]"`."""
            )
        LLMInterfaceV2.__init__(
            self,
            model_name=model_name,
            model_params=model_params or {},
            rate_limit_handler=rate_limit_handler,
            **kwargs,
        )
        self.cohere = cohere
        self.cohere_api_error = cohere.core.api_error.ApiError

        self.client = cohere.ClientV2(**kwargs)
        self.async_client = cohere.AsyncClientV2(**kwargs)

    # overloads for LLMInterface and LLMInterfaceV2 methods
    @overload  # type: ignore[no-overload-impl]
    def invoke(
        self,
        input: str,
        message_history: Optional[Union[List[LLMMessage], MessageHistory]] = None,
        system_instruction: Optional[str] = None,
    ) -> LLMResponse: ...

    @overload
    def invoke(
        self,
        input: List[LLMMessage],
        response_format: Optional[Union[Type[BaseModel], dict[str, Any]]] = None,
        **kwargs: Any,
    ) -> LLMResponse: ...

    @overload  # type: ignore[no-overload-impl]
    async def ainvoke(
        self,
        input: str,
        message_history: Optional[Union[List[LLMMessage], MessageHistory]] = None,
        system_instruction: Optional[str] = None,
    ) -> LLMResponse: ...

    @overload
    async def ainvoke(
        self,
        input: List[LLMMessage],
        response_format: Optional[Union[Type[BaseModel], dict[str, Any]]] = None,
        **kwargs: Any,
    ) -> LLMResponse: ...

    # switching logics to LLMInterface or LLMInterfaceV2
    def invoke(  # type: ignore[no-redef]
        self,
        input: Union[str, List[LLMMessage]],
        message_history: Optional[Union[List[LLMMessage], MessageHistory]] = None,
        system_instruction: Optional[str] = None,
        response_format: Optional[Union[Type[BaseModel], dict[str, Any]]] = None,
        **kwargs: Any,
    ) -> LLMResponse:
        if isinstance(input, str):
            return self.__invoke_v1(input, message_history, system_instruction)
        elif isinstance(input, list):
            return self.__invoke_v2(input, response_format=response_format, **kwargs)
        else:
            raise ValueError(f"Invalid input type for invoke method - {type(input)}")

    async def ainvoke(  # type: ignore[no-redef]
        self,
        input: Union[str, List[LLMMessage]],
        message_history: Optional[Union[List[LLMMessage], MessageHistory]] = None,
        system_instruction: Optional[str] = None,
        response_format: Optional[Union[Type[BaseModel], dict[str, Any]]] = None,
        **kwargs: Any,
    ) -> LLMResponse:
        if isinstance(input, str):
            return await self.__ainvoke_v1(input, message_history, system_instruction)
        elif isinstance(input, list):
            return await self.__ainvoke_v2(
                input, response_format=response_format, **kwargs
            )
        else:
            raise ValueError(f"Invalid input type for ainvoke method - {type(input)}")

    # implementations
    @rate_limit_handler_decorator
    def __invoke_v1(
        self,
        input: str,
        message_history: Optional[Union[List[LLMMessage], MessageHistory]] = None,
        system_instruction: Optional[str] = None,
    ) -> LLMResponse:
        """Sends text to the LLM and returns a response.

        Args:
            input (str): The text to send to the LLM.
            message_history (Optional[Union[List[LLMMessage], MessageHistory]]): A collection previous messages,
                with each message having a specific role assigned.
            system_instruction (Optional[str]): An option to override the llm system message for this invocation.

        Returns:
            LLMResponse: The response from the LLM.
        """
        try:
            if isinstance(message_history, MessageHistory):
                message_history = message_history.messages
            messages = self.get_messages(input, message_history, system_instruction)
            res = self.client.chat(
                messages=messages,
                model=self.model_name,
            )
        except self.cohere_api_error as e:
            raise LLMGenerationError(e)
        return LLMResponse(
            content=res.message.content[0].text if res.message.content else "",  # type: ignore[union-attr]
        )

    @rate_limit_handler_decorator
    def __invoke_v2(
        self,
        input: List[LLMMessage],
        response_format: Optional[Union[Type[BaseModel], dict[str, Any]]] = None,
        **kwargs: Any,
    ) -> LLMResponse:
        """Sends text to the LLM and returns a response.

        Args:
            input (List[LLMMessage]): The messages to send to the LLM.
            response_format: Not supported by CohereLLM.

        Returns:
            LLMResponse: The response from the LLM.
        """
        if response_format is not None:
            raise NotImplementedError(
                "CohereLLM does not currently support structured output"
            )
        try:
            messages = self.get_messages_v2(input)
            res = self.client.chat(
                messages=messages,
                model=self.model_name,
            )
        except self.cohere_api_error as e:
            raise LLMGenerationError("Error calling cohere") from e

        return LLMResponse(
            content=(
                res.message.content[0].text
                if res.message.content and hasattr(res.message.content[0], "text")
                else ""
            ),
        )

    @async_rate_limit_handler_decorator
    async def __ainvoke_v1(
        self,
        input: str,
        message_history: Optional[Union[List[LLMMessage], MessageHistory]] = None,
        system_instruction: Optional[str] = None,
    ) -> LLMResponse:
        """Asynchronously sends text to the LLM and returns a response.

        Args:
            input (str): The text to send to the LLM.
            message_history (Optional[Union[List[LLMMessage], MessageHistory]]): A collection previous messages,
                with each message having a specific role assigned.
            system_instruction (Optional[str]): An option to override the llm system message for this invocation.

        Returns:
            LLMResponse: The response from the LLM.
        """
        try:
            if isinstance(message_history, MessageHistory):
                message_history = message_history.messages
            messages = self.get_messages(input, message_history, system_instruction)
            res = await self.async_client.chat(
                messages=messages,
                model=self.model_name,
            )
        except self.cohere_api_error as e:
            raise LLMGenerationError(e)
        return LLMResponse(
            content=res.message.content[0].text if res.message.content else "",  # type: ignore[union-attr]
        )

    @async_rate_limit_handler_decorator
    async def __ainvoke_v2(
        self,
        input: List[LLMMessage],
        response_format: Optional[Union[Type[BaseModel], dict[str, Any]]] = None,
        **kwargs: Any,
    ) -> LLMResponse:
        if response_format is not None:
            raise NotImplementedError(
                "CohereLLM does not currently support structured output"
            )
        try:
            messages = self.get_messages_v2(input)
            res = await self.async_client.chat(
                messages=messages,
                model=self.model_name,
            )
        except self.cohere_api_error as e:
            raise LLMGenerationError("Error calling cohere") from e
        return LLMResponse(
            content=(
                res.message.content[0].text
                if res.message.content and hasattr(res.message.content[0], "text")
                else ""
            ),
        )

    # subsdiary methods
    def get_messages(
        self,
        input: str,
        message_history: Optional[Union[List[LLMMessage], MessageHistory]] = None,
        system_instruction: Optional[str] = None,
    ) -> ChatMessages:
        """Converts input and message history to ChatMessages for Cohere."""
        messages = []
        if system_instruction:
            messages.append(SystemMessage(content=system_instruction).model_dump())
        if message_history:
            if isinstance(message_history, MessageHistory):
                message_history = message_history.messages
            try:
                MessageList(messages=cast(list[BaseMessage], message_history))
            except ValidationError as e:
                raise LLMGenerationError(e.errors()) from e
            messages.extend(cast(Iterable[dict[str, Any]], message_history))
        messages.append(UserMessage(content=input).model_dump())
        return messages  # type: ignore

    def get_messages_v2(
        self,
        input: list[LLMMessage],
    ) -> ChatMessages:
        """Converts a list of LLMMessage to ChatMessages for Cohere."""
        messages: ChatMessages = []
        for i in input:
            if i["role"] == "system":
                messages.append(self.cohere.SystemChatMessageV2(content=i["content"]))
            elif i["role"] == "user":
                messages.append(self.cohere.UserChatMessageV2(content=i["content"]))
            elif i["role"] == "assistant":
                messages.append(
                    self.cohere.AssistantChatMessageV2(content=i["content"])
                )
            else:
                raise ValueError(f"Unknown role: {i['role']}")
        return messages



#  Copyright (c) "Neo4j"
#  Neo4j Sweden AB [https://neo4j.com]
#  #
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#  #
#      https://www.apache.org/licenses/LICENSE-2.0
#  #
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
"""This example illustrates how to visualize a Pipeline"""

from neo4j_graphrag.experimental.pipeline import Component, Pipeline
from neo4j_graphrag.experimental.pipeline.component import DataModel
from pydantic import validate_call


class IntDataModel(DataModel):
    value: int
    message: str


class Addition(Component):
    async def run(self, a: int, b: int) -> IntDataModel:
        return IntDataModel(value=a + b, message="addition complete")


class Duplicate(Component):
    def __init__(self, factor: int = 2) -> None:
        self.factor = factor

    async def run(self, number: int) -> IntDataModel:
        return IntDataModel(
            value=number * self.factor, message=f"multiplication by {self.factor} done"
        )


class Save(Component):
    @validate_call
    async def run(self, number: IntDataModel) -> IntDataModel:
        return IntDataModel(value=number.value, message="saved")


if __name__ == "__main__":
    pipe = Pipeline()
    pipe.add_component(Duplicate(), "times_two")
    pipe.add_component(Duplicate(factor=10), "times_ten")
    pipe.add_component(Addition(), "addition")
    pipe.add_component(Save(), "save")
    pipe.connect("times_two", "addition", {"a": "times_two.value"})
    pipe.connect("times_ten", "addition", {"b": "times_ten.value"})
    pipe.connect("addition", "save", {"number": "addition"})
    pipe.draw("graph.html")
    pipe.draw("graph_full.html", hide_unused_outputs=False)



"""The LLM interface is compatible with LangChain chat API,
 so any LangChain implementation can be used. Same for embedders.
 For instance, in GraphRAG:

Requires OPENAI_API_KEY to be in the env var.
"""

import neo4j
from langchain_openai.chat_models import ChatOpenAI
from langchain_openai.embeddings import OpenAIEmbeddings
from neo4j_graphrag.generation import GraphRAG
from neo4j_graphrag.retrievers import VectorCypherRetriever

# Define database credentials
URI = "neo4j+s://demo.neo4jlabs.com"
AUTH = ("recommendations", "recommendations")
DATABASE = "recommendations"
INDEX = "moviePlotsEmbedding"


driver = neo4j.GraphDatabase.driver(
    URI,
    auth=AUTH,
)

embedder = OpenAIEmbeddings(model="text-embedding-ada-002")

retriever = VectorCypherRetriever(
    driver,
    index_name=INDEX,
    retrieval_query="WITH node, score RETURN node.title as title, node.plot as plot",
    embedder=embedder,  # type: ignore[arg-type, unused-ignore]
    neo4j_database=DATABASE,
)

llm = ChatOpenAI(model="gpt-5", temperature=0)

rag = GraphRAG(
    retriever=retriever,
    llm=llm,  # type: ignore[arg-type, unused-ignore]
)

result = rag.search(
    "Tell me more about Avatar movies",
    return_context=False,
)
print(result.answer)

driver.close()
